# -*- coding: utf-8 -*-
"""
Generates images from Stable Diffusion
"""

import asyncio
import io
import re
import typing

import discord

from oobabot import fancy_logger
from oobabot import http_client
from oobabot import ooba_client
from oobabot import prompt_generator
from oobabot import sd_client
from oobabot import templates


async def image_task_to_file(image_task: "asyncio.Task[bytes]", image_request: str):
    await image_task
    img_bytes = image_task.result()
    file_of_bytes = io.BytesIO(img_bytes)
    file = discord.File(file_of_bytes)
    file.filename = "photo.png"
    file.description = f"image generated from '{image_request}'"
    return file


class StableDiffusionImageView(discord.ui.View):
    """
    A View that displays buttons to regenerate an image
    from Stable Diffusion with a new seed, or to lock
    in the current image.
    """

    LABEL_ACCEPT = "Accept"
    LABEL_DELETE = "Delete"

    # these two phrases (along with exactly two periods)
    # in "Drawing.." were chosen because they render at
    # the exact same width as each other.  If they don't,
    # the buttons will shift to the left and right as the
    # labels are swapped.
    LABEL_TRY_AGAIN = "Try Again"
    LABEL_DRAWING = "Drawing.."

    def __init__(
        self,
        stable_diffusion_client: sd_client.StableDiffusionClient,
        is_channel_nsfw: bool,
        image_prompt: str,
        requesting_user_id: int,
        requesting_user_name: str,
        template_store: templates.TemplateStore,
    ):
        super().__init__(timeout=120.0)

        self.template_store = template_store

        # only the user who requested generation of the image
        # can have it replaced
        self.requesting_user_id = requesting_user_id
        self.requesting_user_name = requesting_user_name
        self.image_prompt = image_prompt
        self.photo_accepted = False

        #####################################################
        # "Try Again" button
        #
        btn_try_again = discord.ui.Button(
            label=self.LABEL_TRY_AGAIN,
            style=discord.ButtonStyle.blurple,
            row=1,
        )
        self.image_message = None

        async def on_try_again(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return

            try:
                btn_try_again.label = self.LABEL_DRAWING

                # we disable all three buttons because otherwise
                # the lock_in and delete buttons will flicker
                # when we disable the try_again button.  And it
                # doesn't make much sense for them to work anyway
                # when the button is being regenerated.
                btn_try_again.disabled = True
                btn_lock_in.disabled = True
                btn_delete.disabled = True

                await interaction.response.defer()
                await self.get_image_message().edit(view=self)

                # generate a new image
                regen_task = stable_diffusion_client.generate_image(
                    image_prompt, is_channel_nsfw
                )
                regen_file = await image_task_to_file(regen_task, image_prompt)

                btn_try_again.label = self.LABEL_TRY_AGAIN
                btn_try_again.disabled = False
                btn_lock_in.disabled = False
                btn_delete.disabled = False

                await self.get_image_message().edit(attachments=[regen_file], view=self)
            except (http_client.OobaHttpClientError, discord.DiscordException) as err:
                fancy_logger.get().error(
                    "Could not regenerate image: %s", err, exc_info=True
                )

        btn_try_again.callback = on_try_again

        #####################################################
        # "Accept" button
        #
        btn_lock_in = discord.ui.Button(
            label=self.LABEL_ACCEPT,
            style=discord.ButtonStyle.success,
            row=1,
        )

        async def on_lock_in(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return
            await interaction.response.defer()
            await self.detach_view_keep_img()

        btn_lock_in.callback = on_lock_in

        #####################################################
        # "Delete" button
        #
        btn_delete = discord.ui.Button(
            label=self.LABEL_DELETE,
            style=discord.ButtonStyle.danger,
            row=1,
        )

        async def on_delete(interaction: discord.Interaction):
            result = await self.diy_interaction_check(interaction)
            if not result:
                # unauthorized user
                return
            await interaction.response.defer()
            await self.delete_image()

        btn_delete.callback = on_delete

        super().add_item(btn_try_again).add_item(btn_lock_in).add_item(btn_delete)

    def set_image_message(self, image_message: discord.Message):
        self.image_message = image_message

    def get_image_message(self) -> discord.Message:
        if self.image_message is None:
            raise ValueError("image_message is None")
        return self.image_message

    async def delete_image(self):
        await self.detach_view_delete_img(self.get_detach_message())

    async def detach_view_delete_img(self, detach_msg: str):
        await self.get_image_message().edit(
            content=detach_msg,
            view=None,
            attachments=[],
        )

    async def detach_view_keep_img(self):
        self.photo_accepted = True
        await self.get_image_message().edit(
            content=None,
            view=None,
        )

    async def on_timeout(self):
        if not self.photo_accepted:
            await self.delete_image()

    async def diy_interaction_check(self, interaction: discord.Interaction) -> bool:
        """
        Only allow the requesting user to interact with this view.
        """
        if interaction.user.id == self.requesting_user_id:
            return True
        error_message = self._get_message(templates.Templates.IMAGE_UNAUTHORIZED)
        await interaction.response.send_message(
            content=error_message,
            ephemeral=True,
        )
        return False

    def get_image_message_text(self) -> str:
        return self._get_message(templates.Templates.IMAGE_CONFIRMATION)

    def get_detach_message(self) -> str:
        return self._get_message(templates.Templates.IMAGE_DETACH)

    def _get_message(self, message_type: templates.Templates) -> str:
        return self.template_store.format(
            message_type,
            {
                templates.TemplateToken.USER_NAME: self.requesting_user_name,
                templates.TemplateToken.IMAGE_PROMPT: self.image_prompt,
            },
        )


class ImageGenerator:
    """
    Generates images from a given prompt, and posts that image as a
    message to a given channel.
    """

    # if a potential image prompt is shorter than this, we will
    # conclude that it is not an image prompt.
    MIN_IMAGE_PROMPT_LENGTH = 3

    def __init__(
        self,
        ooba_client: ooba_client.OobaClient,
        persona_settings: typing.Dict[str, typing.Any],
        prompt_generator: prompt_generator.PromptGenerator,
        sd_settings: typing.Dict[str, typing.Any],
        stable_diffusion_client: sd_client.StableDiffusionClient,
        template_store: templates.TemplateStore,
    ):
        self.ai_name = persona_settings.get("ai_name", "")
        self.ooba_client = ooba_client
        self.image_words = sd_settings.get("image_words", [])
        self.prompt_generator = prompt_generator
        self.stable_diffusion_client = stable_diffusion_client
        self.template_store = template_store

        self.image_patterns = [
            re.compile(
                r"^.*\b" + image_word + r"\b[\s]*(of|with)?[\s]*[:]?(.*)$",
                re.IGNORECASE,
            )
            for image_word in self.image_words
        ]

    def on_ready(self):
        """
        Called when the bot is connected to Discord.
        """
        fancy_logger.get().debug(
            "Stable Diffusion: image keywords: %s",
            ", ".join(self.image_words),
        )

    # @fancy_logger.log_async_task
    async def _generate_image(
        self,
        image_prompt: str,
        raw_message: discord.Message,
        response_channel: discord.abc.Messageable,
    ) -> discord.Message:
        is_channel_nsfw = False

        # note: public threads in NSFW channels are not considered here
        if isinstance(raw_message.channel, discord.TextChannel):
            is_channel_nsfw = raw_message.channel.is_nsfw()

        image_task = self.stable_diffusion_client.generate_image(
            image_prompt, is_channel_nsfw=is_channel_nsfw
        )
        try:
            file = await image_task_to_file(image_task, image_prompt)
        except (http_client.OobaHttpClientError, discord.DiscordException) as err:
            fancy_logger.get().error("Could not generate image: %s", err, exc_info=True)
            error_message = self.template_store.format(
                templates.Templates.IMAGE_GENERATION_ERROR,
                {
                    templates.TemplateToken.USER_NAME: raw_message.author.display_name,
                    templates.TemplateToken.IMAGE_PROMPT: image_prompt,
                },
            )
            return await response_channel.send(error_message, reference=raw_message)

        regen_view = StableDiffusionImageView(
            self.stable_diffusion_client,
            is_channel_nsfw=is_channel_nsfw,
            image_prompt=image_prompt,
            requesting_user_id=raw_message.author.id,
            requesting_user_name=raw_message.author.display_name,
            template_store=self.template_store,
        )

        kwargs = {}
        # we can only pass a reference if the message is in the same channel
        # as the original request.  Also, send() won't take None of this
        # argument, so we need to conditionally add it.
        if raw_message.channel == response_channel:
            kwargs["reference"] = raw_message

        image_message = await response_channel.send(
            content=regen_view.get_image_message_text(),
            file=file,
            view=regen_view,
            **kwargs,
        )
        regen_view.image_message = image_message
        return image_message

    def maybe_get_image_prompt(
        self, raw_message: discord.Message
    ) -> typing.Optional[str]:
        for image_pattern in self.image_patterns:
            match = image_pattern.search(raw_message.content)
            if match:
                image_prompt = match.group(2)
                if len(image_prompt) < self.MIN_IMAGE_PROMPT_LENGTH:
                    continue
                fancy_logger.get().debug("Found image prompt: %s", image_prompt)
                return image_prompt
        return None

    async def generate_image(
        self,
        user_image_keywords: str,
        raw_message: discord.Message,
        response_channel: discord.abc.Messageable,
    ) -> "asyncio.Task[discord.Message]":
        """
        Kick off a task to generate an image, post it to the channel,
        and return message the image is posted in.
        """
        return asyncio.create_task(
            self._generate_image(user_image_keywords, raw_message, response_channel)
        )



# -*- coding: utf-8 -*-
"""
A replacement voice client for discord.py, using the Discrivener audio transcriber.
"""

import asyncio
import typing

import discord
import discord.backoff
import discord.gateway
import discord.guild
import discord.opus
import discord.state
import discord.types
from discord.types import voice  # this is so pylint doesn't complain

from oobabot import audio_responder
from oobabot import discrivener
from oobabot import discrivener_message
from oobabot import fancy_logger
from oobabot import ooba_client  # pylint: disable=unused-import
from oobabot import prompt_generator  # pylint: disable=unused-import
from oobabot import transcript
from oobabot import types


class VoiceClientError(Exception):
    """
    Base exception class for voice client errors.
    """


class VoiceClient(discord.VoiceProtocol):
    """
    A replacement voice client for discord.py, using the Discrivener audio transcriber.


    You do not create these, you typically get them from
    e.g. :meth:`VoiceChannel.connect`, by passing in
    cls=voice_client.VoiceClient
    """

    discrivener_location: str
    discrivener_model_location: str
    current_instance: typing.Optional["VoiceClient"] = None
    ooba_client: ooba_client.OobaClient
    prompt_generator: prompt_generator.PromptGenerator
    wakewords: typing.List[str] = []

    supported_modes: typing.Tuple[voice.SupportedModes, ...] = (
        "xsalsa20_poly1305",
        "xsalsa20_poly1305_suffix",
        "xsalsa20_poly1305_lite",
    )

    def __init__(
        self,
        client: discord.client.Client,
        channel: discord.abc.Connectable,
    ) -> None:
        super().__init__(client, channel)

        if not isinstance(channel, discord.VoiceChannel):
            raise ValueError("Channel is not a voice channel.")

        if channel.guild is None:
            raise ValueError("Channel does not have a guild.")

        if client.user is None:
            raise ValueError("Client does not have a user.")

        self._discrivener = discrivener.Discrivener(
            self.discrivener_location,
            self.discrivener_model_location,
            self._handle_discrivener_output,
        )
        self._discrivener_connected = False
        self._handshaking = False
        self._oobabot_voice_connected = False
        self._potentially_reconnecting = False
        self._voice_state_complete = asyncio.Event()
        self._voice_server_complete = asyncio.Event()
        self._state: discord.state.ConnectionState = client._connection
        self._session_id = discord.utils.MISSING
        self._server_id = discord.utils.MISSING
        self._transcript = transcript.Transcript(client.user.id, self.wakewords)
        self._guild_channel = channel
        self._user = client.user

        self._audio_responder = audio_responder.AudioResponder(
            channel,
            self._discrivener,
            self.ooba_client,
            self.prompt_generator,
            self._transcript,
        )

    @property
    def guild(self) -> discord.guild.Guild:
        """
        :class:`Guild`: The guild we're connected to.
        """
        return self._guild_channel.guild

    @property
    def user(self) -> discord.user.ClientUser:
        """
        :class:`ClientUser`: The user connected to voice (i.e. ourselves).
        """
        return self._user

    @property
    def session_id(self) -> str:
        """
        :class:`str`: The session ID for this voice connection.
        """
        return self._session_id

    async def on_voice_state_update(
        self,
        data: voice.GuildVoiceState,
        /,
    ) -> None:
        self._session_id = data["session_id"]
        channel_id = data["channel_id"]

        if not self._handshaking or self._potentially_reconnecting:
            fancy_logger.get().debug(
                "Server-initiated voice state update for Channel ID %s (Guild ID %s)",
                channel_id,
                self.guild.id,
            )

            # If we're done handshaking then we just need to update ourselves
            # If we're potentially reconnecting due to a 4014, then we need to
            # differentiate
            # a channel move and an actual force disconnect
            if channel_id is None:
                # We're being disconnected so cleanup
                await self.disconnect()
            else:
                channel = self.guild.get_channel(int(channel_id))
                if channel is None:
                    fancy_logger.get().warning(
                        "Channel ID %s not found in Guild ID %s",
                        channel_id,
                        self.guild.id,
                    )
                    return
                if not isinstance(channel, discord.VoiceChannel):
                    fancy_logger.get().warning(
                        "Channel ID %s not a VoiceChannel.", channel_id
                    )
                    return
                self.channel = channel
        else:
            fancy_logger.get().debug(
                "Voice state complete for Channel ID %s (Guild ID %s) during handshake",
                channel_id,
                self.guild.id,
            )
            self._voice_state_complete.set()

    async def on_voice_server_update(
        self,
        data: voice.VoiceServerUpdate,
        /,
    ) -> None:
        if self._voice_server_complete.is_set():
            fancy_logger.get().warning("Ignoring extraneous voice server update.")
            return

        token = data["token"]
        guild_id = int(data["guild_id"])
        endpoint = data.get("endpoint")

        if endpoint is None or token is None:
            fancy_logger.get().warning(
                "Awaiting endpoint... This requires waiting. "
                "If timeout occurred considering raising the timeout and reconnecting."
            )
            return

        endpoint, _, _ = endpoint.rpartition(":")
        if endpoint.startswith("wss://"):
            # Just in case, strip it off since we're going to add it later
            endpoint = endpoint[6:]

        # this will start the process, we'll get a callback when the connection
        # is made
        await self._discrivener.run(
            self._guild_channel.id,
            endpoint,
            guild_id,
            self.session_id,
            self.user.id,
            token,
        )
        await self._audio_responder.start()
        self._voice_server_complete.set()

    async def voice_connect(
        self, self_deaf: bool = False, self_mute: bool = False
    ) -> None:
        await self._guild_channel.guild.change_voice_state(
            channel=self._guild_channel, self_deaf=self_deaf, self_mute=self_mute
        )

    async def voice_disconnect(self) -> None:
        fancy_logger.get().info(
            "The voice handshake is being terminated for Channel ID %s (Guild ID %s)",
            self._guild_channel.id,
            self.guild.id,
        )
        self._oobabot_voice_connected = False
        await self._guild_channel.guild.change_voice_state(channel=None)
        await self._discrivener.stop()
        await self._audio_responder.stop()
        VoiceClient.current_instance = None

    async def connect(
        self,
        *,
        reconnect: bool,
        timeout: float,
        self_deaf: bool = False,
        self_mute: bool = False,
    ) -> None:
        fancy_logger.get().info("Connecting to voice...")

        if self.is_connected():
            raise VoiceClientError("Already connected to a voice channel.")

        self._voice_state_complete.clear()
        self._voice_server_complete.clear()

        futures = [
            self._voice_state_complete.wait(),
            self._voice_server_complete.wait(),
        ]

        self._handshaking = True

        fancy_logger.get().info("Starting voice handshake...")

        await self.voice_connect(self_deaf, self_mute)

        try:
            await discord.utils.sane_wait_for(futures, timeout=timeout)
        except asyncio.TimeoutError as err:
            await self.disconnect(force=True)
            raise VoiceClientError(
                f"Couldn't connect to voice channel within {timeout:.2f}s"
            ) from err

        self._oobabot_voice_connected = True

        self._voice_state_complete.clear()
        self._voice_server_complete.clear()

        fancy_logger.get().info("Voice handshake complete.")
        self._handshaking = False
        VoiceClient.current_instance = self

    async def potential_reconnect(self) -> bool:
        # Attempt to stop the player thread from playing early
        # self._potentially_reconnecting = True
        fancy_logger.get().warning("voice_client::potential_reconnect: not implemented")
        return False

    async def disconnect(self, *, force: bool = False) -> None:
        """|coro|

        Disconnects this voice client from voice.
        """
        if not force and not self.is_connected():
            return

        try:
            await self.voice_disconnect()
        finally:
            self.cleanup()

    async def move_to(self, channel: typing.Optional[discord.abc.Snowflake]) -> None:
        """|coro|

        Moves you to a different voice channel.

        Parameters
        -----------
        channel: Optional[:class:`abc.Snowflake`]
            The channel to move to. Must be a voice channel.
        """
        # todo: tell songbird to move channels
        fancy_logger.get().warning("voice_client::move_to: not implemented")
        await self._guild_channel.guild.change_voice_state(channel=channel)

    def is_connected(self) -> bool:
        """Indicates if the voice client is connected to voice."""
        if not self._oobabot_voice_connected:
            return False
        if self._discrivener is None:
            return False
        if not self._discrivener.is_running():
            return False
        return self._discrivener_connected

    def _handle_discrivener_output(self, message: "types.DiscrivenerMessage"):
        if message.type in (
            types.DiscrivenerMessageType.CONNECT,
            types.DiscrivenerMessageType.RECONNECT,
        ):
            fancy_logger.get().debug("discrivener message: %s", message)
            self._voice_state_complete.set()
            self._discrivener_connected = True

        elif types.DiscrivenerMessageType.DISCONNECT == message.type:
            fancy_logger.get().debug("discrivener message: %s", message)
            self._voice_state_complete.set()
            self._discrivener_connected = False

            # we're disconnected, so have the voice_client disconnect
            # too... however we can't call disconnect() here because
            # we would create a deadlock.  The disconnect() method
            # will terminate the process, then wait for its notification
            # threads to exit... and we're on the notification thread
            # right now.  So to break the loop, use asyncio to schedule
            # the disconnect() call to happen later.
            loop = asyncio.get_event_loop()
            if loop is None:
                fancy_logger.get().warning(
                    "No event loop to schedule voice_client.disconnect() call"
                )
                return
            # todo: how to wait for a result here?
            loop.call_soon_threadsafe(self.disconnect)

        elif types.DiscrivenerMessageType.USER_JOIN == message.type:
            fancy_logger.get().debug("discrivener message: %s", message)

        elif types.DiscrivenerMessageType.USER_LEAVE == message.type:
            fancy_logger.get().debug("discrivener message: %s", message)

        elif types.DiscrivenerMessageType.TRANSCRIPTION == message.type:
            if isinstance(message, discrivener_message.UserVoiceMessage):
                self._transcript.on_transcription(message)
            else:
                fancy_logger.get().warning(
                    "Unexpected message value %s", type(message).__name__
                )

        elif types.DiscrivenerMessageType.CHANNEL_SILENT == message.type:
            if isinstance(message, discrivener_message.ChannelSilentData):
                self._transcript.on_channel_silent(message)
            else:
                fancy_logger.get().warning(
                    "Unexpected message value %s", type(message).__name__
                )

        else:
            fancy_logger.get().warning(
                "Unknown discrivener message type: %s", message.type
            )

    def current_transcript(self) -> typing.Optional[transcript.Transcript]:
        return self._transcript



# -*- coding: utf-8 -*-
"""
Converts Discord library objects into generic objects that can be used by the AI

This is done to make it easier to swap out the Discord library for something else
in the future, and to make it easier to test the AI without having to mock the
Discord library.
"""


import base64
import functools
import os
import pathlib
import re
import typing

import discord

from oobabot import fancy_logger
from oobabot import types

FORBIDDEN_CHARACTERS = r"[\n\r\t]"
FORBIDDEN_CHARACTERS_PATTERN = re.compile(FORBIDDEN_CHARACTERS)


def get_channel_name(channel: discord.abc.Messageable) -> str:
    if isinstance(channel, discord.Thread):
        return "thread #" + channel.name
    if isinstance(channel, discord.abc.GuildChannel):
        return "channel #" + channel.name
    if isinstance(channel, discord.DMChannel):
        return "-DM-"
    if isinstance(channel, discord.GroupChannel):
        return "-GROUP-DM-"
    return "-Unknown-"


def sanitize_string(raw_string: str) -> str:
    """
    Filter out any characters that would confuse the AI
    """
    return FORBIDDEN_CHARACTERS_PATTERN.sub(" ", raw_string)


def discord_message_to_generic_message(
    raw_message: discord.Message,
) -> typing.Union[types.GenericMessage, types.ChannelMessage, types.DirectMessage]:
    """
    Convert a discord message to a GenericMessage or subclass thereof
    """
    generic_args = {
        "author_id": raw_message.author.id,
        "author_name": sanitize_string(raw_message.author.display_name),
        "channel_id": raw_message.channel.id,
        "channel_name": get_channel_name(raw_message.channel),
        "message_id": raw_message.id,
        "body_text": sanitize_string(raw_message.content),
        "author_is_bot": raw_message.author.bot,
        "send_timestamp": raw_message.created_at.timestamp(),
        "reference_message_id": raw_message.reference.message_id
        if raw_message.reference
        else "",
    }
    if isinstance(raw_message.channel, discord.DMChannel):
        return types.DirectMessage(**generic_args)
    if isinstance(
        raw_message.channel,
        (
            discord.TextChannel,
            discord.GroupChannel,
            discord.Thread,
            discord.VoiceChannel,
        ),
    ):
        return types.ChannelMessage(
            mentions=[mention.id for mention in raw_message.mentions],
            **generic_args,
        )
    fancy_logger.get().warning(
        f"Unknown channel type {type(raw_message.channel)}, "
        + f"unsolicited replies disabled.: {raw_message.channel}"
    )
    return types.GenericMessage(**generic_args)


def replace_mention_ids_with_names(
    generic_message: types.GenericMessage,
    fn_user_id_to_name: typing.Callable[["re.Match[str]"], str],
):
    """
    Replace user ID mentions with the user's chosen display
    name in the given guild (aka server)
    """
    # it looks like normal IDs are 18 digits.  But give it some
    # wiggle room in case things change in the future.
    # e.g.: <@009999999999999999>
    at_mention_pattern = r"<@(\d{16,20})>"
    while True:
        match = re.search(at_mention_pattern, generic_message.body_text)
        if not match:
            break
        generic_message.body_text = (
            generic_message.body_text[: match.start()]
            + fn_user_id_to_name(match)
            + generic_message.body_text[match.end() :]
        )


def dm_user_id_to_name(
    bot_user_id: int,
    bot_name: str,
) -> typing.Callable[["re.Match[str]"], str]:
    """
    Replace user ID mentions with the bot's name.  Used when
    we are in a DM with the bot.
    """
    if " " in bot_name:
        bot_name = f'"{bot_name}"'

    def _replace_user_id_mention(match: typing.Match[str]) -> str:
        user_id = int(match.group(1))
        print(f"bot_user_id={bot_user_id}, user_id={user_id}")
        if user_id == bot_user_id:
            return f"@{bot_name}"
        return match.group(0)

    return _replace_user_id_mention


def guild_user_id_to_name(
    guild: discord.Guild,
) -> typing.Callable[["re.Match[str]"], str]:
    def _replace_user_id_mention(match: typing.Match[str]) -> str:
        user_id = int(match.group(1))
        member = guild.get_member(user_id)
        if member is None:
            return match.group(0)
        display_name = member.display_name
        if " " in display_name:
            display_name = f'"{display_name}"'
        return f"@{display_name}"

    return _replace_user_id_mention


def get_intents() -> discord.Intents:
    intents = discord.Intents.default()
    intents.message_content = True
    intents.members = True
    return intents


async def test_discord_token(discord_token: str) -> bool:
    class _SimplestBot(discord.Client):
        async def on_ready(self):
            self.has_connected = True
            await self.close()

        def __init__(self):
            super().__init__(intents=get_intents())
            self.has_connected = False

    simplest_bot = _SimplestBot()
    try:
        await simplest_bot.start(discord_token, reconnect=False)
    except discord.errors.ConnectionClosed as err:
        # in theory, discord.errors.PrivilegedIntentsRequired
        # should get fired in this case, but it doesn't
        if err.code != 4014:
            raise
        fancy_logger.get().warning(
            "The bot token you provided does not have the required "
            + "gateway intents.  Did you remember to enable both "
            + "'SERVER MEMBERS INTENT' and 'MESSAGE CONTENT INTENT' "
            + "in the bot's settings on Discord?",
        )
        return False

    except discord.LoginFailure:
        return False
    finally:
        await simplest_bot.close()
    return simplest_bot.has_connected


def get_user_id_from_token(discord_token: str) -> int:
    """
    Extract the bot's user ID from the discord token.
    """

    # turns out, the discord_token includes our client ID, so we can just
    # extract it from there.
    #
    # the discord token has this format:
    # AAAAAAAAAAAAAAAAAAAAAAAAAA.BBBBBB.CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
    #
    # where each section, A, B, and C, is independently a base64-encoded string.
    #
    # Section A encodes the bot's client ID, which is the decimal encoding of
    # a 64-bit number in the range [21154535154122752, 18446744073709551615]
    # (17 to 20 digits long).
    #
    # The other sections aren't important here.
    token_parts = discord_token.split(".")
    token_part_a = token_parts[0]

    # the base64 decoder requires the string to be a multiple of 4 characters
    # long, so we need to add padding
    if len(token_part_a) % 4 != 0:
        token_part_a += "=" * (4 - len(token_part_a) % 4)

    return int(base64.b64decode(token_part_a).decode("utf-8"))


def generate_invite_url(ai_user_id: int) -> str:
    # we want to generate a URL like this:
    # https://discord.com/api/oauth2/authorize?client_id={client_id}&permissions={permissions}}&scope=bot
    #
    # where {client_id} is the bot's client ID, and {permissions} is the
    # permissions bit array with our desired permissions set.

    # for the permissions bit array, we can generate it with the library
    permissions = discord.Permissions(
        change_nickname=True,
        send_messages=True,
        create_public_threads=True,
        send_messages_in_threads=True,
        attach_files=True,
        read_message_history=True,
        read_messages=True,
        add_reactions=True,
    ).value

    return (
        "https://discord.com/api/oauth2/authorize?client_id="
        + f"{ai_user_id}&permissions={permissions}&scope=bot"
    )


def setup_logging(**kwargs: typing.Any):
    discord.utils.setup_logging(**kwargs)


async def fail_interaction(
    interaction: discord.Interaction, reason: typing.Optional[str] = None
):
    command = "unknown command"
    if interaction.command is not None:
        command = interaction.command.name

    if reason is None:
        reason = f"{command} failed"

    fancy_logger.get().warning(
        "interaction failed: command='%s', user='%s', channel='%s', reason='%s'",
        command,
        interaction.user,
        interaction.channel,
        reason,
    )

    await interaction.response.send_message(reason, ephemeral=True, silent=True)


def _file_exists_and_is_file(filepath: typing.Optional[str]) -> typing.Optional[str]:
    if filepath is None:
        return None

    path = pathlib.Path(filepath).expanduser()
    if not path.is_file():
        return None

    if not os.access(path, os.R_OK):
        return None

    return str(path.resolve())


def validate_discrivener_locations(
    discrivener_location: str,
    discrivener_model_location: str,
) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
    """
    Verify that the file discrivener_location exists
    and is a file.

    If that passes, also checks that discrivener_model_location
    exists and is a file.

    Returns the expanded paths to each files if it exists,
    or None if it doesn't.

    Returns the tuple of these checks for the two passed files.
    """
    actual_discrivener_location = _file_exists_and_is_file(discrivener_location)
    # check that the discrivener binary is executable as well
    if actual_discrivener_location is not None:
        if not os.access(actual_discrivener_location, os.X_OK):
            fancy_logger.get().warning(
                "discrivener binary is not executable: %s",
                actual_discrivener_location,
            )
            actual_discrivener_location = None

    actual_model_location = _file_exists_and_is_file(discrivener_model_location)
    return (actual_discrivener_location, actual_model_location)


# the following class was modified from O'Reilly's Python Cookbook,
# chapter 5, section 19.  Its use is allowed under this license:
# Copyright (c) 2001, Sébastien Keim
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
#    * Neither the name of the <ORGANIZATION> nor the names of its
#      contributors may be used to endorse or promote products derived
#      from this software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

T = typing.TypeVar("T")
S = typing.TypeVar("S")


class RingBuffer(typing.Generic[T]):
    """
    A generic ring buffer.
    """

    def __init__(self, size_max: int):
        self.cur = 0
        self.max = size_max
        self.data: typing.List[T] = []

    class _FullRingBuffer(typing.Generic[S]):
        """
        Class implementing the RingBuffer when it's full.
        With python class magic, this class is swapped in when the
        buffer becomes full.
        """

        cur: int
        max: int
        data: typing.List[S]

        def append(self, val: S) -> None:
            """
            Append an element overwriting the oldest one.
            """
            self.data[self.cur] = val
            self.cur = (self.cur + 1) % self.max

        def get(self) -> typing.List[S]:
            """
            Return a list of elements from the oldest to the newest.
            """
            return self.data[self.cur :] + self.data[: self.cur]

        def size(self) -> int:
            """
            Return the size of the buffer.
            """
            return self.max

    def append(self, val: T) -> None:
        """
        Append an element at the end of the buffer.
        """
        self.data.append(val)
        if len(self.data) == self.max:
            self.cur = 0
            # Permanently change self's class from non-full to full
            self.__class__ = self._FullRingBuffer

    def get(self) -> typing.List[T]:
        """
        Return a list of elements from the oldest to the newest.
        """
        return self.data

    def size(self) -> int:
        """
        Return the number of elements currently in the buffer.
        """
        return len(self.data)


# end of O'Reilly code


@functools.lru_cache
def author_from_user_id(
    user_id: int,
    guild: discord.Guild,
) -> typing.Optional["types.FancyAuthor"]:
    member = guild.get_member(user_id)
    if member is None:
        return None
    if member.avatar:
        avatar_url = member.avatar.url
    else:
        avatar_url = None
    if member.accent_color:
        accent_color = member.accent_color.to_rgb()
    else:
        accent_color = (0, 0, 0)
    return types.FancyAuthor(
        user_id=user_id,
        author_is_bot=member.bot,
        author_name=member.display_name,
        author_accent_color=accent_color,
        author_avatar_url=avatar_url,
    )



# -*- coding: utf-8 -*-
"""
Main bot class.  Contains Discord-specific code that can't
be easily extracted into a cross-platform library.
"""

import asyncio
import typing

import discord

from oobabot import bot_commands
from oobabot import decide_to_respond
from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import image_generator
from oobabot import ooba_client
from oobabot import persona
from oobabot import prompt_generator
from oobabot import repetition_tracker
from oobabot import response_stats
from oobabot import types


class DiscordBot(discord.Client):
    """
    Main bot class.  Connects to Discord, monitors for messages,
    and dispatches responses.
    """

    def __init__(
        self,
        bot_commands: bot_commands.BotCommands,
        decide_to_respond: decide_to_respond.DecideToRespond,
        discord_settings: dict,
        image_generator: typing.Optional[image_generator.ImageGenerator],
        ooba_client: ooba_client.OobaClient,
        persona: persona.Persona,
        prompt_generator: prompt_generator.PromptGenerator,
        repetition_tracker: repetition_tracker.RepetitionTracker,
        response_stats: response_stats.AggregateResponseStats,
    ):
        self.bot_commands = bot_commands
        self.decide_to_respond = decide_to_respond
        self.image_generator = image_generator
        self.ooba_client = ooba_client
        self.persona = persona
        self.prompt_generator = prompt_generator
        self.repetition_tracker = repetition_tracker
        self.response_stats = response_stats

        self.ai_user_id = -1

        self.dont_split_responses = discord_settings["dont_split_responses"]
        self.ignore_dms = discord_settings["ignore_dms"]
        self.reply_in_thread = discord_settings["reply_in_thread"]
        self.stop_markers = discord_settings["stop_markers"]
        self.stream_responses = discord_settings["stream_responses"]
        self.stream_responses_speed_limit = discord_settings[
            "stream_responses_speed_limit"
        ]

        # add stopping_strings to stop_markers
        self.stop_markers.extend(self.ooba_client.get_stopping_strings())

        super().__init__(intents=discord_utils.get_intents())

    async def on_ready(self) -> None:
        guilds = self.guilds
        num_guilds = len(guilds)
        num_channels = sum(len(guild.channels) for guild in guilds)

        if self.user:
            self.ai_user_id = self.user.id
            user_id_str = self.user.name
        else:
            user_id_str = "<unknown>"

        fancy_logger.get().info(
            "Connected to discord as %s (ID: %d)", user_id_str, self.ai_user_id
        )
        fancy_logger.get().debug(
            "monitoring %d channels across %d server(s)", num_channels, num_guilds
        )
        if self.ignore_dms:
            fancy_logger.get().debug("Ignoring DMs")
        else:
            fancy_logger.get().debug("listening to DMs")

        if self.stream_responses:
            fancy_logger.get().debug(
                "Response Grouping: streamed live into a single message"
            )
        elif self.dont_split_responses:
            fancy_logger.get().debug("Response Grouping: returned as single messages")
        else:
            fancy_logger.get().debug(
                "Response Grouping: split into messages by sentence"
            )

        fancy_logger.get().debug("AI name: %s", self.persona.ai_name)
        fancy_logger.get().debug("AI persona: %s", self.persona.persona)

        fancy_logger.get().debug(
            "History: %d lines ", self.prompt_generator.history_lines
        )

        fancy_logger.get().debug(
            "Stop markers: %s", ", ".join(self.stop_markers) or "<none>"
        )

        # log unsolicited_channel_cap
        cap = self.decide_to_respond.get_unsolicited_channel_cap()
        cap = str(cap) if cap > 0 else "<unlimited>"
        fancy_logger.get().debug(
            "Unsolicited channel cap: %s",
            cap,
        )

        str_wakewords = (
            ", ".join(self.persona.wakewords) if self.persona.wakewords else "<none>"
        )
        fancy_logger.get().debug("Wakewords: %s", str_wakewords)

        self.ooba_client.on_ready()

        if self.image_generator is None:
            fancy_logger.get().debug("Stable Diffusion: disabled")
        else:
            self.image_generator.on_ready()

        # we do this at the very end because when you restart
        # the bot, it can take a while for the commands to
        # register
        try:
            # register the commands
            await self.bot_commands.on_ready(self)
        except discord.DiscordException as err:
            fancy_logger.get().warning(
                "Failed to register commands: %s (continuing without commands)", err
            )

        # show a warning if the bot is connected to zero guilds,
        # with a helpful link on how to fix it
        if num_guilds == 0:
            fancy_logger.get().warning(
                "The bot is not connected to any servers.  "
                + "Please add the bot to a server here:",
            )
            fancy_logger.get().warning(
                discord_utils.generate_invite_url(self.ai_user_id)
            )

    async def on_message(self, raw_message: discord.Message) -> None:
        """
        Called when a message is received from Discord.

        This method is called for every message that the bot can see.
        It decides whether to respond to the message, and if so,
        calls _handle_response() to generate a response.

        :param raw_message: The raw message from Discord.
        """
        try:
            message = discord_utils.discord_message_to_generic_message(raw_message)
            should_respond, is_summon = self.decide_to_respond.should_reply_to_message(
                self.ai_user_id, message
            )
            if not should_respond:
                return

            is_summon_in_public_channel = is_summon and isinstance(
                message,
                types.ChannelMessage,
            )

            async with raw_message.channel.typing():
                await self._handle_response(
                    message, raw_message, is_summon_in_public_channel
                )

        except discord.DiscordException as err:
            fancy_logger.get().error(
                "Exception while processing message: %s", err, exc_info=True
            )

    async def _handle_response(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        is_summon_in_public_channel: bool,
    ) -> None:
        """
        Called when we've decided to respond to a message.

        It decides if we're sending a text response, an image response,
        or both, and then sends the response(s).
        """
        image_prompt = None
        if self.image_generator is not None:
            # are we creating an image?
            image_prompt = self.image_generator.maybe_get_image_prompt(raw_message)

        result = await self._send_text_response(
            message=message,
            raw_message=raw_message,
            image_requested=image_prompt is not None,
            is_summon_in_public_channel=is_summon_in_public_channel,
        )
        if result is None:
            # we failed to create a thread that the user could
            # read our response in, so we're done here.  Abort!
            return
        message_task, response_channel = result

        # log the mention, now that we know the channel we
        # want to monitor later to continue to conversation
        if isinstance(response_channel, (discord.Thread, discord.abc.GuildChannel)):
            if is_summon_in_public_channel:
                self.decide_to_respond.log_mention(
                    response_channel.id,
                    message.send_timestamp,
                )

        image_task = None
        if self.image_generator is not None and image_prompt is not None:
            image_task = await self.image_generator.generate_image(
                image_prompt,
                raw_message,
                response_channel=response_channel,
            )

        response_tasks = [
            task for task in [message_task, image_task] if task is not None
        ]
        await asyncio.wait(response_tasks)

        # there may be more than one exception, so be sure to log
        # them all before raising either of them
        raise_later = None
        for task in response_tasks:
            if task.exception() is not None:
                fancy_logger.get().error(
                    f"Exception while running {task.get_coro()} "
                    + f"response: {task.exception()}",
                    stack_info=True,
                )
                raise_later = task.exception()
        if raise_later is not None:
            raise raise_later

    async def _send_text_response(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        image_requested: bool,
        is_summon_in_public_channel: bool,
    ) -> typing.Optional[typing.Tuple[asyncio.Task, discord.abc.Messageable]]:
        """
        Send a text response to a message.

        This method determines what channel or thread to post the message
        in, creating a thread if necessary.  It then posts the message
        by calling _send_text_response_to_channel().

        Returns a tuple of the task that was created to send the message,
        and the channel that the message was sent to.

        If no message was sent, the task and channel will be None.
        """
        response_channel = raw_message.channel
        if (
            self.reply_in_thread
            and isinstance(raw_message.channel, discord.TextChannel)
            and isinstance(raw_message.author, discord.Member)
        ):
            # we want to create a response thread, if possible
            # but we have to see if the user has permission to do so
            # if the user can't we wont respond at all.
            perms = raw_message.channel.permissions_for(raw_message.author)
            if perms.create_public_threads:
                response_channel = await raw_message.create_thread(
                    name=f"{self.persona.ai_name}, replying to "
                    + f"{raw_message.author.display_name}",
                )
                fancy_logger.get().debug(
                    f"Created response thread {response_channel.name} "
                    + f"({response_channel.id}) "
                    + f"in {raw_message.channel.name}"
                )
            else:
                # This user can't create threads, so we won't respond.
                # The reason we don't respond in the channel is that
                # it can create confusion later if a second user who
                # DOES have thread-create permission replies to that
                # message.  We'd end up creating a thread for that
                # second user's response, and again for a third user,
                # etc.
                fancy_logger.get().debug("User can't create threads, not responding.")
                return None

        response_coro = self._send_text_response_in_channel(
            message=message,
            raw_message=raw_message,
            image_requested=image_requested,
            is_summon_in_public_channel=is_summon_in_public_channel,
            response_channel=response_channel,
            response_channel_id=response_channel.id,
        )
        response_task = asyncio.create_task(response_coro)
        return (response_task, response_channel)

    async def _send_text_response_in_channel(
        self,
        message: types.GenericMessage,
        raw_message: discord.Message,
        image_requested: bool,
        is_summon_in_public_channel: bool,
        response_channel: discord.abc.Messageable,
        response_channel_id: int,
    ) -> None:
        """
        Getting closer now!  This method is what actually gathers message
        history, queries the AI for a text response, breaks the response
        into individual messages, and then and then calls
        __send_response_message() to send sech message.
        """
        fancy_logger.get().debug(
            "Request from %s in %s", message.author_name, message.channel_name
        )

        repeated_id = self.repetition_tracker.get_throttle_message_id(
            response_channel_id
        )

        # determine if we're responding to a specific message that
        # summoned us.  If so, find out what message ID that was, so
        # that we can ignore all messages sent after it (as not to
        # confuse the AI about what to reply to)
        reference = None
        ignore_all_until_message_id = None
        if is_summon_in_public_channel:
            # we can't use the message reference if we're starting a new thread
            if message.channel_id == response_channel_id:
                reference = raw_message.to_reference()
            ignore_all_until_message_id = raw_message.id

        recent_messages = await self._recent_messages_following_thread(
            channel=response_channel,
            num_history_lines=self.prompt_generator.history_lines,
            stop_before_message_id=repeated_id,
            ignore_all_until_message_id=ignore_all_until_message_id,
        )

        prompt_prefix = await self.prompt_generator.generate(
            message_history=recent_messages,
            image_requested=image_requested,
        )

        this_response_stat = self.response_stats.log_request_arrived(prompt_prefix)

        # restrict the @mentions the AI is allowed to use in its response.
        # this is to prevent another user from being able to trick the AI
        # into @-pinging a large group and annoying them.
        # Only the author of the original message may be @-pinged.
        allowed_mentions = discord.AllowedMentions(
            everyone=False,
            users=[raw_message.author],
            roles=False,
        )

        # will be set to true when we abort the response because:
        #  it was empty
        #  it repeated a previous response and we're throttling it
        aborted_by_us = False
        sent_message_count = 0
        try:
            if self.stream_responses:
                generator = self.ooba_client.request_as_grouped_tokens(
                    prompt_prefix, interval=self.stream_responses_speed_limit
                )
                last_sent_message = await self._render_streaming_response(
                    generator,
                    this_response_stat,
                    response_channel,
                    response_channel_id,
                    allowed_mentions,
                    reference,
                )
                if last_sent_message is not None:
                    sent_message_count = 1
            else:
                if self.dont_split_responses:
                    response = await self.ooba_client.request_as_string(prompt_prefix)
                    (
                        last_sent_message,
                        aborted_by_us,
                    ) = await self._send_response_message(
                        response,
                        this_response_stat,
                        response_channel,
                        response_channel_id,
                        allowed_mentions,
                        reference,
                    )
                    if last_sent_message is not None:
                        sent_message_count = 1
                else:
                    sent_message_count = 0
                    last_sent_message = None
                    async for sentence in self.ooba_client.request_by_message(
                        prompt_prefix
                    ):
                        (
                            sent_message,
                            abort_response,
                        ) = await self._send_response_message(
                            sentence,
                            this_response_stat,
                            response_channel,
                            response_channel_id,
                            allowed_mentions=allowed_mentions,
                            reference=reference,
                        )
                        if sent_message is not None:
                            last_sent_message = sent_message
                            sent_message_count += 1
                            # only use the reference for the first
                            # message in a multi-message chain
                            reference = None
                        if abort_response:
                            aborted_by_us = True
                            break

        except discord.DiscordException as err:
            fancy_logger.get().error("Error: %s", err, exc_info=True)
            self.response_stats.log_response_failure()
            return

        if 0 == sent_message_count:
            if aborted_by_us:
                fancy_logger.get().warning(
                    "No response sent.  The AI has generated a message that we have "
                    + "chosen not to send, probably because it was empty or repeated."
                )
            else:
                fancy_logger.get().warning(
                    "An empty response was received from Oobabooga.  Please check that "
                    + "the AI is running properly on the Oobabooga server at %s.",
                    self.ooba_client.base_url,
                )
            self.response_stats.log_response_failure()
            return

        this_response_stat.write_to_log(f"Response to {message.author_name} done!  ")
        self.response_stats.log_response_success(this_response_stat)

    async def _send_response_message(
        self,
        response: str,
        this_response_stat: response_stats.ResponseStats,
        response_channel: discord.abc.Messageable,
        response_channel_id: int,
        allowed_mentions: discord.AllowedMentions,
        reference: typing.Optional[discord.MessageReference],
    ) -> typing.Tuple[typing.Optional[discord.Message], bool]:
        """
        Given a string that represents an individual response message,
        post it in the given channel.

        It also looks to see if a message contains a termination string,
        and if so it will return False to indicate that we should stop
        the response.

        Also does some bookkeeping to make sure we don't repeat ourselves,
        and to track how many messages we've sent.

        Returns a tuple with:
        - the sent discord message, if any
        - a boolean indicating if we need to abort the response entirely
        """
        (sentence, abort_response) = self._filter_immersion_breaking_lines(response)
        if abort_response:
            return (None, True)
        if not sentence:
            # we can't send an empty message
            return (None, False)

        response_message = await response_channel.send(
            sentence,
            allowed_mentions=allowed_mentions,
            suppress_embeds=True,
            reference=reference,  # type: ignore
        )
        self.repetition_tracker.log_message(
            response_channel_id,
            discord_utils.discord_message_to_generic_message(response_message),
        )

        this_response_stat.log_response_part()
        return (response_message, False)

    async def _render_streaming_response(
        self,
        response_iterator: typing.AsyncIterator[str],
        this_response_stat: response_stats.ResponseStats,
        response_channel: discord.abc.Messageable,
        response_channel_id: int,
        allowed_mentions: discord.AllowedMentions,
        reference: typing.Optional[discord.MessageReference],
    ) -> typing.Optional[discord.Message]:
        response = ""
        last_message = None
        async for token in response_iterator:
            if "" == token:
                continue

            response += token
            (response, abort_response) = self._filter_immersion_breaking_lines(response)

            # if we are aborting a response, we want to at least post
            # the valid parts, so don't abort quite yet.

            if last_message is None:
                if not response:
                    # we don't want to send an empty message
                    continue

                # when we send the first message, we don't want to send a notification,
                # as it will only include the first token of the response.  This will
                # not be very useful to anyone.
                last_message = await response_channel.send(
                    response,
                    allowed_mentions=allowed_mentions,
                    silent=True,
                    suppress_embeds=True,
                    reference=reference,  # type: ignore
                )
            else:
                await last_message.edit(
                    content=response,
                    allowed_mentions=allowed_mentions,
                    suppress=True,
                )
                last_message.content = response

            # we want to abort the response only after we've sent any valid
            # messages, and potentially removed any partial immersion-breaking
            # lines that we posted when they were in the process of being received.
            if abort_response:
                break

            this_response_stat.log_response_part()

        if last_message is not None:
            self.repetition_tracker.log_message(
                response_channel_id,
                discord_utils.discord_message_to_generic_message(last_message),
            )

        return last_message

    def _filter_immersion_breaking_lines(
        self, sentence: str
    ) -> typing.Tuple[str, bool]:
        """
        Given a string that represents an individual response message,
        filter out any lines that would break immersion.

        These include lines that include a termination symbol, lines
        that attempt to carry on the conversation as a different user,
        and lines that include text which is part of the AI prompt.

        Returns the subset of the input string that should be sent,
        and a boolean indicating if we should abort the response entirely,
        ignoring any further lines.
        """
        lines = sentence.split("\n")
        good_lines = []
        previous_line = ""
        abort_response = False
        for line in lines:
            # if the AI gives itself a second line, just ignore
            # the line instruction and continue
            if self.prompt_generator.bot_prompt_line == line:
                fancy_logger.get().warning(
                    "Filtered out %s from response, continuing", line
                )
                continue

            # hack: abort response if it looks like the AI is
            # continuing the conversation as someone else
            if line.endswith(" says:"):
                fancy_logger.get().warning(
                    'Filtered out "%s" from response, aborting', line
                )
                abort_response = True
                break

            # look for partial stop markers within a line
            for marker in self.stop_markers:
                if marker in line:
                    (keep_part, removed) = line.split(marker, 1)
                    fancy_logger.get().warning(
                        'Filtered out "%s" from response, aborting',
                        removed,
                    )
                    if keep_part:
                        good_lines.append(keep_part)
                    abort_response = True
                    break

            if not line and not previous_line:
                # filter out multiple blank lines in a row
                continue

            if not line.strip():
                # filter out lines that are entirely made of whitespace
                continue

            good_lines.append(line)
        return ("\n".join(good_lines), abort_response)

    ########
    async def _filter_history_message(
        self,
        message: discord.Message,
        stop_before_message_id: typing.Optional[int],
    ) -> typing.Tuple[typing.Optional[types.GenericMessage], bool]:
        """
        Filter out any messages that we don't want to include in the
        AI's history.

        These include:
         - messages generated by our image generator
         - messages at or before the stop_before_message_id

        Also, modify the message in the following ways:
         - if the message is from the AI, set the author name to
           the AI's persona name, not its Discord account name
         - remove <@_0000000_> user-id based message mention text,
           replacing them with @username mentions
        """
        # if we've hit the throttle message, stop and don't add any
        # more history
        if stop_before_message_id and message.id == stop_before_message_id:
            return (None, False)

        generic_message = discord_utils.discord_message_to_generic_message(message)

        if generic_message.author_id == self.ai_user_id:
            # make sure the AI always sees its persona name
            # in the transcript, even if the chat program
            # has it under a different account name
            generic_message.author_name = self.persona.ai_name

            # hack: use the suppress_embeds=True flag to indicate
            # that this message is one we generated as part of a text
            # response, as opposed to an image or application message
            if not message.flags.suppress_embeds:
                # this is a message generated by our image generator
                return (None, True)

        if message.channel.guild is None:
            fn_user_id_to_name = discord_utils.dm_user_id_to_name(
                self.ai_user_id,
                self.persona.ai_name,
            )
        else:
            fn_user_id_to_name = discord_utils.guild_user_id_to_name(
                message.channel.guild,
            )

        discord_utils.replace_mention_ids_with_names(
            generic_message,
            fn_user_id_to_name=fn_user_id_to_name,
        )
        return (generic_message, True)

    async def _filtered_history_iterator(
        self,
        async_iter_history: typing.AsyncIterator[discord.Message],
        stop_before_message_id: typing.Optional[int],
        ignore_all_until_message_id: typing.Optional[int],
        limit: int,
    ) -> typing.AsyncIterator[types.GenericMessage]:
        """
        When returning the history of a thread, Discord
        does not include the message that kicked off the thread.

        It will show it in the UI as if it were, but it's not
        one of the messages returned by the history iterator.

        This method attempts to return that message as well,
        if we need it.
        """
        items = 0
        last_returned = None
        ignoring_all = ignore_all_until_message_id is not None
        async for item in async_iter_history:
            if items >= limit:
                return

            if ignoring_all:
                if item.id == ignore_all_until_message_id:
                    ignoring_all = False
                else:
                    # this message was sent after the message we're
                    # responding to.  So filter out it as to not confuse
                    # the AI into responding to content from that message
                    # instead
                    continue

            last_returned = item
            (sanitized_message, allow_more) = await self._filter_history_message(
                item,
                stop_before_message_id=stop_before_message_id,
            )
            if not allow_more:
                # we've hit a message which requires us to stop
                # and look at more history
                return
            if sanitized_message is not None:
                yield sanitized_message
                items += 1

        if last_returned is not None and items < limit:
            # we've reached the beginning of the history, but
            # still have space.  If this message was a reply
            # to another message, return that message as well.
            if last_returned.reference is None:
                return

            ref = last_returned.reference.resolved

            # the resolved message may be None if the message
            # was deleted
            if ref is not None and isinstance(ref, discord.Message):
                (sanitized_message, _) = await self._filter_history_message(
                    ref,
                    stop_before_message_id,
                )
                if sanitized_message is not None:
                    yield sanitized_message

    # when looking through the history of a channel, we'll have a goal
    # of retrieving a certain number of lines of history.  However,
    # there are some messages in the history that we'll want to filter
    # out.  These include messages that were generated by our image
    # generator, as well as certain messages that will be ignored
    # in order to generate a response for a specific user who
    # @-mentions the bot.
    #
    # This is the maximum number of "extra" messages to retrieve
    # from the history, in an attempt to find enough messages
    # that we can filter out the ones we don't want and still
    # have enough left over to satisfy the request.
    #
    # Note that since the history is returned in reverse order,
    # and each is pulled in only as needed, there's not much of a
    # penalty to making this somewhat large.  But still, we want
    # to keep it reasonable.
    MESSAGE_HISTORY_LOOKBACK_BONUS = 20

    async def _recent_messages_following_thread(
        self,
        channel: discord.abc.Messageable,
        stop_before_message_id: typing.Optional[int],
        ignore_all_until_message_id: typing.Optional[int],
        num_history_lines: int,
    ) -> typing.AsyncIterator[types.GenericMessage]:
        max_messages_to_check = num_history_lines + self.MESSAGE_HISTORY_LOOKBACK_BONUS
        history = channel.history(limit=max_messages_to_check)
        result = self._filtered_history_iterator(
            history,
            limit=num_history_lines,
            stop_before_message_id=stop_before_message_id,
            ignore_all_until_message_id=ignore_all_until_message_id,
        )
        return result



# -*- coding: utf-8 -*-
"""
Detects when the bot is repeating previous messages, and attempts
to fix this by hiding the messages that it's repeating from its view
of the chat history.

Is also used to implement the /lobotomize command, which is the same
thing except it's triggered by a command instead of automatically.
"""
import typing

from oobabot import fancy_logger
from oobabot import types


class RepetitionTracker:
    """
    Tracks the last message the bot posted in each channel, and
    the number of times in a row it has been repeated.
    """

    def __init__(self, repetition_threshold: int) -> None:
        self.repetition_threshold = repetition_threshold

        # stores a map of channel_id ->
        #   (last_message, throttle_message_id, repetition_count)

        self.repetition_count: typing.Dict[int, typing.Tuple[str, int, int]] = {}

    def get_throttle_message_id(self, channel_id: int) -> int:
        """
        Returns the message ID of the last message that should be throttled, or 0
        if no throttling is needed
        """
        _, throttle_message_id, _ = self.repetition_count.get(
            channel_id, (None, 0, None)
        )
        return throttle_message_id

    def log_message(
        self, channel_id: int, response_message: types.GenericMessage
    ) -> None:
        """
        Logs a message sent by the bot, to be used for repetition tracking
        """
        # make string into canonical form
        sentence = self.make_canonical(response_message.body_text)

        last_message, throttle_message_id, repetition_count = self.repetition_count.get(
            channel_id, ("", 0, 0)
        )
        if last_message == sentence:
            repetition_count += 1
        else:
            repetition_count = 0

        if repetition_count > 0:
            fancy_logger.get().debug(
                "Repetition count for channel %d is %d", channel_id, repetition_count
            )

        if self.should_throttle(repetition_count):
            fancy_logger.get().warning(
                "Repetition found, will throttle history for channel #%d "
                + "in next request",
                channel_id,
            )
            throttle_message_id = response_message.message_id

        self.repetition_count[channel_id] = (
            sentence,
            throttle_message_id,
            repetition_count,
        )

    def hide_messages_before(self, channel_id: int, message_id: int) -> None:
        """
        Hides all messages before the given message ID in the given channel
        """
        sentence, _, repetition_count = self.repetition_count.get(
            channel_id, ("", 0, 0)
        )
        fancy_logger.get().info(
            "Hiding messages before message ID %d in channel %d", message_id, channel_id
        )
        self.repetition_count[channel_id] = (sentence, message_id, repetition_count)

    def should_throttle(self, repetition_count: int) -> bool:
        """
        Returns whether the bot should throttle history for a given repetition count
        """
        return repetition_count >= self.repetition_threshold

    def make_canonical(self, content: str) -> str:
        return content.strip().lower()



# -*- coding: utf-8 -*-
"""
Discrivener process launcher and handler.  Receives and
parses messages from the Discrivener process and passes
them to the handler.
"""

import asyncio
import json
import pathlib
import signal
import typing

from oobabot import discrivener_message
from oobabot import fancy_logger
from oobabot import types


class Discrivener:
    """
    Launches and handles the Discrivener process.
    """

    KILL_TIMEOUT: float = 2.0

    # pylint: disable=R1732
    def __init__(
        self,
        discrivener_location: str,
        discrivener_model_location: str,
        handler: typing.Callable[["types.DiscrivenerMessage"], None],
        log_file: typing.Optional[str] = None,
    ):
        self._discrivener_location = pathlib.Path(discrivener_location).expanduser()
        self._discrivener_model_location = pathlib.Path(
            discrivener_model_location
        ).expanduser()
        self._handler: typing.Callable[["types.DiscrivenerMessage"], None] = handler
        self._process: typing.Optional["asyncio.subprocess.Process"] = None
        self._stderr_reading_task: typing.Optional[asyncio.Task] = None
        self._stdout_reading_task: typing.Optional[asyncio.Task] = None
        if log_file is not None:
            self._log_file = open(log_file, "a", encoding="utf-8")
        else:
            self._log_file = None

    # pylint: enable=R1732

    async def run(
        self,
        channel_id: int,
        endpoint: str,
        guild_id: int,
        session_id: str,
        user_id: int,
        voice_token: str,
    ):
        if self.is_running():
            raise RuntimeError("Already running")

        args = (
            "--channel-id",
            str(channel_id),
            "--endpoint",
            endpoint,
            "--guild-id",
            str(guild_id),
            "--session-id",
            session_id,
            "--user-id",
            str(user_id),
            "--voice-token",
            voice_token,
            str(self._discrivener_model_location),
        )
        await self._launch_process(args)

    async def stop(self):
        if self.is_running():
            await self._kill_process()

    def is_running(self):
        return self._process is not None

    async def _launch_process(self, args: typing.Tuple[str, ...]):
        fancy_logger.get().info(
            "Launching Discrivener process: %s", self._discrivener_location
        )
        fancy_logger.get().debug(
            "Using Discrivener model file: %s", self._discrivener_model_location
        )

        self._process = await asyncio.create_subprocess_exec(
            self._discrivener_location,
            *args,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        fancy_logger.get().info(
            "Discrivener process started, PID: %d", self._process.pid
        )

        self._stderr_reading_task = asyncio.create_task(self._read_stderr())
        self._stdout_reading_task = asyncio.create_task(self._read_stdout())

    async def _kill_process(self):
        if self._process is None:
            return
        self._process.send_signal(signal.SIGINT)
        try:
            await asyncio.wait_for(self._process.wait(), timeout=self.KILL_TIMEOUT)
            fancy_logger.get().info(
                "Discrivener process (PID %d) stopped gracefully", self._process.pid
            )

        except asyncio.TimeoutError:
            fancy_logger.get().warning(
                "Discrivener process (PID %d) did not exit after %d seconds, killing",
                self._process.pid,
                self.KILL_TIMEOUT,
            )
            self._process.kill()
            await asyncio.wait_for(self._process.wait(), timeout=self.KILL_TIMEOUT)
            fancy_logger.get().warning(
                "Discrivener process (PID %d) force-killed", self._process.pid
            )
        finally:
            self._process = None
            if self._stderr_reading_task is not None:
                self._stderr_reading_task.cancel()
            if self._stdout_reading_task is not None:
                self._stdout_reading_task.cancel()

        # terminate stdout and stderr reading tasks
        if self._stderr_reading_task is not None:
            await asyncio.wait_for(self._stderr_reading_task, timeout=self.KILL_TIMEOUT)
            self._stderr_reading_task = None

        if self._stdout_reading_task is not None:
            await asyncio.wait_for(self._stdout_reading_task, timeout=self.KILL_TIMEOUT)
            self._stdout_reading_task = None

    # @fancy_logger.log_async_task
    async def _read_stdout(self):
        while True:
            try:
                if self._process is None or self._process.stdout is None:
                    fancy_logger.get().debug(
                        "Discrivener stdout reader: _process went away, exiting"
                    )
                    break
                line_bytes = await self._process.stdout.readuntil()
            except asyncio.IncompleteReadError:
                break
            line = line_bytes.decode("utf-8").strip()

            if self._log_file is not None:
                try:
                    self._log_file.write(line + "\n")
                except (IOError, OSError) as err:
                    fancy_logger.get().warning(
                        "transcript: failed to log to file: %s", err
                    )
            try:
                message = json.loads(
                    line,
                    object_pairs_hook=discrivener_message.object_pairs_hook,
                )
                self._handler(message)
            except json.JSONDecodeError:
                fancy_logger.get().error("Discrivener: could not parse %s", line)

        fancy_logger.get().info("Discrivener stdout reader exited")

    # @fancy_logger.log_async_task
    async def _read_stderr(self):
        print("reading stderr")
        # loop until EOF, printing everything to stderr
        while True:
            try:
                if self._process is None or self._process.stderr is None:
                    fancy_logger.get().debug(
                        "Discrivener stderr reader: _process went away, exiting"
                    )
                    break
                line_bytes = await self._process.stderr.readuntil()
            except asyncio.IncompleteReadError:
                break
            line = line_bytes.decode("utf-8").strip()
            if (
                "whisper_init_state: " in line
                or "whisper_init_from_file_no_state: " in line
                or "whisper_model_load: " in line
            ):
                # workaround nonsense noise in whisper.cpp
                continue
            fancy_logger.get().error("Discrivener: %s", line)

        fancy_logger.get().info("Discrivener stderr reader exited")

    def speak(self, text: str):
        if self._process is None or self._process.stdin is None:
            fancy_logger.get().error("Discrivener: _process is not running")
            return

        self._process.stdin.write(text.encode("utf-8") + b"\n")



# -*- coding: utf-8 -*-
"""
Generic types for messages, used to abstract away
Discord-specific types from the rest of the code.
"""
import abc
import datetime
import enum
import typing


class GenericMessage:
    """
    Represents a message from a user.
    """

    def __init__(
        self,
        author_id: int,
        author_name: str,
        channel_id: int,
        channel_name: str,
        message_id: int,
        reference_message_id,
        body_text: str,
        author_is_bot: bool,
        send_timestamp: float,
    ):
        self.author_id = author_id
        self.author_name = author_name
        self.message_id = message_id
        self.body_text = body_text
        self.author_is_bot = author_is_bot
        self.reference_message_id = reference_message_id
        self.send_timestamp = send_timestamp
        self.channel_id = channel_id
        self.channel_name = channel_name

    def is_empty(self) -> bool:
        return not self.body_text.strip()


class DirectMessage(GenericMessage):
    """
    Represents a message sent directly to the bot.
    """


class ChannelMessage(GenericMessage):
    """
    Represents a message sent in a channel, including
    a private group chat or thread.
    """

    def __init__(
        self,
        /,
        mentions: typing.List[int],
        **kwargs,
    ):
        super().__init__(**kwargs)  # type: ignore
        self.mentions = mentions

    def is_mentioned(self, user_id: int) -> bool:
        return user_id in self.mentions


class FancyAuthor:
    """
    Display information about the author of a message.
    """

    def __init__(
        self,
        user_id: int,
        author_is_bot: bool,
        author_name: str,
        author_accent_color: typing.Tuple[int, int, int],
        author_avatar_url: typing.Optional[str],
    ):
        self._user_id = user_id
        self._author_is_bot = author_is_bot
        self._author_name = author_name
        self._author_accent_color = author_accent_color
        self._author_avatar_url = author_avatar_url

    @property
    def user_id(self) -> int:
        """
        Returns the ID of the user who sent the message.
        """
        return self._user_id

    @property
    def author_is_bot(self) -> bool:
        """
        Returns whether the user who sent the message is a bot.
        """
        return self._author_is_bot

    @property
    def author_name(self) -> str:
        """
        Returns the name of the user who sent the message.
        """
        return self._author_name

    @property
    def author_accent_color(self) -> typing.Tuple[int, int, int]:
        """
        Returns the accent color of the user who sent the message.
        """
        return self._author_accent_color

    @property
    def author_avatar_url(self) -> typing.Optional[str]:
        """
        Returns the avatar URL of the user who sent the message,
        if the user has chosen one.

        Will be null if the user has not set an avatar.
        """
        return self._author_avatar_url


class DiscrivenerMessageType(str, enum.Enum):
    """
    Enumerates the different types of Discrivener messages.
    """

    CHANNEL_SILENT = "ChannelSilent"
    CONNECT = "Connect"
    DISCONNECT = "Disconnect"
    RECONNECT = "Reconnect"
    TRANSCRIPTION = "Transcription"
    USER_JOIN = "UserJoin"
    USER_LEAVE = "UserLeave"


class DiscrivenerMessage(abc.ABC):
    """
    Base class for all Discrivener messages.
    """

    @abc.abstractmethod
    def __init__(self, message: dict):
        ...

    type: DiscrivenerMessageType


# not an actual GenericMessage, but still a message
class VoiceMessage(DiscrivenerMessage):
    """
    Represents a message that we have transcribed
    from a voice channel, attributed to a user.
    """

    def __init__(
        self,
        user_id: int,
        start_time: datetime.datetime,
        duration: datetime.timedelta,
    ):
        self._user_id = user_id
        self._start_time = start_time
        self._audio_duration = duration

    @property
    def user_id(self) -> int:
        """
        Returns the user ID of the user who sent the message.
        """
        return self._user_id

    @property
    def start_time(self) -> datetime.datetime:
        """
        Returns the start time of the transcription.
        """
        return self._start_time

    @property
    def duration(self) -> datetime.timedelta:
        """
        Returns the duration of the transcription.
        """
        return self._audio_duration

    @property
    @abc.abstractmethod
    def text(self) -> str:
        """
        Returns the text of the transcription.
        """

    @property
    @abc.abstractmethod
    def is_bot(self) -> bool:
        """
        Returns whether the user who sent the message is a bot.
        """


class VoiceMessageWithTokens(VoiceMessage):
    """
    A voice message that can be broken down into tokens.
    """

    @property
    @abc.abstractmethod
    def tokens_with_confidence(self) -> typing.List[typing.Tuple[str, int]]:
        """
        Returns the tokens of the transcription, with confidence scores.
        """



# -*- coding: utf-8 -*-
"""
Watches the transcript, when a wakeword is detected,
it builds a prompt for the AI, queries it, and
then queues a response.
"""

import asyncio
import typing

import discord

from oobabot import discord_utils
from oobabot import discrivener
from oobabot import fancy_logger
from oobabot import ooba_client
from oobabot import prompt_generator
from oobabot import transcript
from oobabot import types


class AudioResponder:
    """
    Watches the transcript, when a wakeword is detected,
    it builds a prompt for the AI, queries it, and
    then queues a response.
    """

    TASK_TIMEOUT_SECONDS = 5.0

    def __init__(
        self,
        channel: discord.VoiceChannel,
        discrivener: discrivener.Discrivener,
        ooba_client: ooba_client.OobaClient,
        prompt_generator: prompt_generator.PromptGenerator,
        transcript: transcript.Transcript,
    ):
        self._abort = False
        self._channel = channel
        self._discrivener = discrivener
        self._ooba_client = ooba_client
        self._prompt_generator = prompt_generator
        self._transcript = transcript
        self._task: typing.Optional[asyncio.Task] = None

    async def start(self):
        await self.stop()
        self._task = asyncio.create_task(self._transcript_reply_task())

    async def stop(self):
        if self._task is None:
            return

        self._abort = True
        self._task.cancel()
        try:
            await asyncio.wait_for(self._task, timeout=self.TASK_TIMEOUT_SECONDS)
        except asyncio.TimeoutError:
            fancy_logger.get().warning("audio_responder: task did not quit in time")
        except asyncio.CancelledError:
            fancy_logger.get().info("audio_responder: task stopped")
        self._task = None
        self._abort = False

    # @fancy_logger.log_async_task
    async def _transcript_reply_task(self):
        fancy_logger.get().info("audio_responder: started")
        while not self._abort:
            await self._transcript.wakeword_event.wait()
            self._transcript.wakeword_event.clear()
            await self._respond()

        fancy_logger.get().info("audio_responder: exiting")

    async def _respond(self):
        transcript_history = self._transcript_history_iterator()
        prompt_prefix = await self._prompt_generator.generate(
            message_history=transcript_history,
            image_requested=False,
        )

        response = await self._ooba_client.request_as_string(prompt_prefix)
        fancy_logger.get().debug("received response: '%s'", response)

        # wait for silence before responding
        await self._transcript.silence_event.wait()

        # shove response into history
        self._transcript.on_bot_response(response)

        self._discrivener.speak(response)
        # todo: make this a setting?
        # await self._channel.send(response)

    def _transcript_history_iterator(
        self,
    ) -> typing.AsyncIterator[types.GenericMessage]:
        voice_messages = self._transcript.message_buffer.get()
        voice_messages.sort(key=lambda message: message.start_time, reverse=True)

        # create an async generator which iterates over the lines
        # in the transcript
        async def _gen():
            for message in voice_messages:
                author = discord_utils.author_from_user_id(
                    message.user_id,
                    self._channel.guild,
                )
                if author is None:
                    author_name = f"user #{message.user_id}"
                    author_is_bot = message.is_bot
                else:
                    author_name = author.author_name
                    author_is_bot = author.author_is_bot
                yield types.GenericMessage(
                    author_id=message.user_id,
                    author_name=author_name,
                    channel_id=0,
                    channel_name="",
                    message_id=0,
                    reference_message_id=0,
                    body_text=message.text,
                    author_is_bot=author_is_bot,
                    send_timestamp=message.start_time.timestamp(),
                )

        return _gen()



# -*- coding: utf-8 -*-
"""
Stores a transcript of a voice channel.
"""
import asyncio
import datetime
import random
import re
import typing

from oobabot import discord_utils
from oobabot import discrivener_message
from oobabot import fancy_logger
from oobabot import types


class Transcript:
    """
    Stores a transcript of a voice channel.
    """

    NUM_LINES = 300

    def __init__(
        self,
        bot_user_id: int,
        wakewords: typing.List[str],
    ):
        self._bot_user_id = bot_user_id
        self._wakewords: typing.Set[str] = set(word.lower() for word in wakewords)

        self.message_buffer = discord_utils.RingBuffer[types.VoiceMessage](
            self.NUM_LINES
        )
        self.silence_event = asyncio.Event()
        self.wakeword_event = asyncio.Event()
        self.last_mention = datetime.datetime.min

    def on_bot_response(self, text: str):
        """
        Adds a bot response to the transcript.
        """
        self.message_buffer.append(BotVoiceMessage(self._bot_user_id, text))

    def on_transcription(
        self,
        message: discrivener_message.UserVoiceMessage,
    ) -> None:
        self.message_buffer.append(message)

        # todo: what about wakewords which span segments?
        wakeword_found = False
        for word in re.split(r"[ .,!?\"']", message.text):
            if word.lower() in self._wakewords:
                wakeword_found = True
                break

        now = datetime.datetime.now()
        if wakeword_found:
            fancy_logger.get().info("transcript: wakeword detected! %s", message.text)
            self.last_mention = now
            self.wakeword_event.set()
        else:
            # chance of replying within 5 minutes of last reply
            seconds_since_mention = (now - self.last_mention).seconds
            if seconds_since_mention > 2 * 60:
                chance = 0.05
            else:
                chance = 0.95 ** (1 - (seconds_since_mention / 60))
            # also, divide chance by number of human speakers
            # in the message history
            humans = set()
            for msg in self.message_buffer.get():
                if not msg.is_bot:
                    humans.add(msg.user_id)
            if 1 == len(humans):
                chance = 1.0
            else:
                chance /= 3 * len(humans)
            fancy_logger.get().debug(
                "transcript: chance of replying: %f (+seconds: %d, humans: %d)",
                chance,
                seconds_since_mention,
                len(humans),
            )
            if chance > 0.0 and chance > random.random():
                self.wakeword_event.set()

    def on_channel_silent(
        self, activity: discrivener_message.ChannelSilentData
    ) -> None:
        if activity.silent:
            self.silence_event.set()
        else:
            self.silence_event.clear()


class BotVoiceMessage(types.VoiceMessage):
    """
    Represents a fake "transcribed" message generated by
    the bot.  This isn't a real transcription, because we got
    it from the bot, not from Discrivener.  But we're creating
    a similar object to store it in, so that we can use similar
    code to store and display it.
    """

    def __init__(
        self,
        bot_user_id: int,
        text: str,
    ):
        self._text = text
        super().__init__(
            user_id=bot_user_id,
            start_time=datetime.datetime.now(),
            duration=datetime.timedelta(seconds=1),
        )

    @property
    def text(self) -> str:
        return self._text

    @property
    def is_bot(self) -> bool:
        """
        Returns whether the user is a bot.
        """
        return True



# -*- coding: utf-8 -*-
"""
Data classes for all Discrivener messages, plus code
to deserialize them from JSON.
"""


import collections
import datetime
import typing

from oobabot import types


def object_pairs_hook(
    pairs: typing.List[typing.Tuple[str, typing.Any]]
) -> typing.Union["types.DiscrivenerMessage", dict]:
    cls = MESSAGE_TYPE_TO_CLASS.get(pairs[0][0])
    if cls is not None and len(pairs) == 1:
        return cls(pairs[0][1])
    return collections.OrderedDict(pairs)


class ChannelSilentData(types.DiscrivenerMessage):
    """
    Represents whether any user is speaking in the channel.
    """

    def __init__(self, data: dict):
        self.type = types.DiscrivenerMessageType.CHANNEL_SILENT
        self.silent = bool(data)

    def __repr__(self):
        return f"ChannelSilent(silent={self.silent})"


class ConnectData(types.DiscrivenerMessage):
    """
    Represents us connecting or reconnecting to the voice channel.
    """

    def __init__(self, data: dict):
        self.type = types.DiscrivenerMessageType.CONNECT
        self.channel_id = data.get("channel_id")
        self.guild_id = data.get("guild_id")
        self.session_id = data.get("session_id")
        self.server = data.get("server")
        self.ssrc = data.get("ssrc")

    def __repr__(self):
        return (
            f"ConnectData(channel_id={self.channel_id}, "
            + f"guild_id={self.guild_id}, "
            + f"session_id={self.session_id}, "
            + f"server={self.server}, "
            + f"ssrc={self.ssrc})"
        )


class DisconnectData(types.DiscrivenerMessage):
    """
    Represents a disconnect from the voice channel.
    """

    def __init__(self, data: dict):
        self.type = types.DiscrivenerMessageType.DISCONNECT
        self.kind: str = data.get("kind", "unknown")
        self.reason: str = data.get("reason", "unknown")
        self.channel_id: int = data.get("channel_id", 0)
        self.guild_id: int = data.get("guild_id", 0)
        self.session_id: int = data.get("session_id", 0)

    def __repr__(self):
        return (
            f"DisconnectData(kind={self.kind}, "
            + f"reason={self.reason}, "
            + f"channel_id={self.channel_id}, "
            + f"guild_id={self.guild_id}, "
            + f"session_id={self.session_id})"
        )


class UserJoinData(types.DiscrivenerMessage):
    """
    Represents a user joining a voice channel.
    """

    def __init__(self, data: int):
        print(f"UserJoinData data is {data}")
        self.type = types.DiscrivenerMessageType.USER_JOIN
        self.user_id: int = data

    def __str__(self):
        return f"User #{self.user_id} joined voice channel"


class UserLeaveData(types.DiscrivenerMessage):
    """
    Represents a user leaving a voice channel.
    """

    def __init__(self, data: int):
        print(f"UserLeaveData data is {data}")
        self.type = types.DiscrivenerMessageType.USER_LEAVE
        self.user_id: int = data

    def __str__(self):
        return f"User #{self.user_id} left voice channel"


class TokenWithProbability:
    """
    Represents a token with a probability.
    """

    def __init__(self, data: dict):
        self.probability: int = data.get("p", 0)
        self.token_id: int = data.get("token_id", 0)
        self.token_text: str = str(data.get("token_text"))

    def __repr__(self):
        return (
            "TokenWithProbability("
            + f"probability={self.probability}, "
            + f"token_id={self.token_id}, "
            + f"token_text={self.token_text})"
        )


def to_datetime(message: dict) -> datetime.datetime:
    """
    Converts a message into a datetime object.
    """
    seconds: int = message.get("secs_since_epoch", 0)
    nanos: int = message.get("nanos_since_epoch", 0)
    return datetime.datetime.fromtimestamp(seconds + nanos / 1e9)


def to_duration(message: dict) -> datetime.timedelta:
    """
    Converts a message into a timedelta object.
    """
    seconds: int = message.get("secs", 0)
    nanos: int = message.get("nanos", 0)
    return datetime.timedelta(seconds=seconds, microseconds=nanos / 1e3)


class TextSegment:
    """
    Represents a single text segment of a transcribed message.
    """

    def __init__(self, message: dict):
        self.tokens_with_probability = [
            TokenWithProbability(data)
            for data in message.get("tokens_with_probability", [])
        ]
        self.start_offset_ms: int = message.get("start_offset_ms", 0)
        self.end_offset_ms: int = message.get("end_offset_ms", self.start_offset_ms + 1)

    def __repr__(self):
        return (
            f"TextSegment(tokens_with_probability={self.tokens_with_probability}, "
            + f"start_offset_ms={self.start_offset_ms}, "
            + f"end_offset_ms={self.end_offset_ms})"
        )

    def __str__(self) -> str:
        return "".join([t.token_text for t in self.tokens_with_probability])


class UserVoiceMessage(types.VoiceMessageWithTokens):
    """
    Represents a transcribed message.
    """

    def __init__(self, message: dict):
        self.type = types.DiscrivenerMessageType.TRANSCRIPTION
        self._processing_time: datetime.timedelta = to_duration(
            message.get("processing_time", datetime.timedelta(milliseconds=1.0))
        )
        self._segments: typing.List[TextSegment] = [
            TextSegment(s) for s in message.get("segments", [])
        ]
        super().__init__(
            message.get("user_id", 0),
            to_datetime(message.get("start_timestamp", datetime.datetime.now)),
            to_duration(
                message.get("audio_duration", datetime.timedelta(milliseconds=1.0))
            ),
        )
        self._latency: datetime.timedelta = (
            datetime.datetime.now() - self._start_time - self._audio_duration
        )

    @property
    def processing_time(self):
        """
        Returns the processing time of the transcription.
        """
        return self._processing_time

    @property
    def latency(self):
        """
        Returns the latency of the transcription.
        """
        return self._latency

    @property
    def text(self) -> str:
        """
        Returns the text of the transcription.
        """
        return "".join([str(s) for s in self._segments])

    @property
    def is_bot(self) -> bool:
        """
        Returns whether the user is a bot.
        """
        return False

    @property
    def tokens_with_confidence(self) -> typing.List[typing.Tuple[str, int]]:
        """
        Returns the tokens with their confidence.
        """
        return [
            (t.token_text, t.probability)
            for s in self._segments
            for t in s.tokens_with_probability
        ]

    def __repr__(self) -> str:
        return (
            "UserVoiceMessage("
            + f"start_time={self._start_time}, "
            + f"user_id={self._user_id}, "
            + f"audio_duration={self._audio_duration}, "
            + f"processing_time={self._processing_time}, "
            + f"segments={self._segments}, "
            + f"latency={self._latency})"
        )


MESSAGE_TYPE_TO_CLASS: typing.Dict[str, typing.Type[types.DiscrivenerMessage]] = {
    types.DiscrivenerMessageType.CHANNEL_SILENT: ChannelSilentData,
    types.DiscrivenerMessageType.CONNECT: ConnectData,
    types.DiscrivenerMessageType.DISCONNECT: DisconnectData,
    types.DiscrivenerMessageType.RECONNECT: ConnectData,
    types.DiscrivenerMessageType.TRANSCRIPTION: UserVoiceMessage,
    types.DiscrivenerMessageType.USER_JOIN: UserJoinData,
    types.DiscrivenerMessageType.USER_LEAVE: UserLeaveData,
}



# -*- coding: utf-8 -*-
"""
Implementation of commands to join and leave voice channels.
"""
import typing

import discord

from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import ooba_client
from oobabot import persona
from oobabot import prompt_generator
from oobabot import voice_client


class AudioCommands:
    """
    Implementation of commands to join and leave voice channels.
    """

    def __init__(
        self,
        persona: persona.Persona,
        ooba_client: ooba_client.OobaClient,
        prompt_generator: prompt_generator.PromptGenerator,
        discrivener_location: str,
        discrivener_model_location: str,
    ):
        voice_client.VoiceClient.wakewords = persona.wakewords

        self.discrivener_location = discrivener_location
        self.persona = persona
        self.voice_client: typing.Optional[voice_client.VoiceClient] = None

        voice_client.VoiceClient.discrivener_location = discrivener_location
        voice_client.VoiceClient.discrivener_model_location = discrivener_model_location
        voice_client.VoiceClient.ooba_client = ooba_client
        voice_client.VoiceClient.prompt_generator = prompt_generator

    def _discover_voice_channel(
        self, interaction: discord.Interaction
    ) -> typing.Optional[discord.VoiceChannel]:
        if isinstance(interaction.user, discord.Member):
            # if invoked from a guild channel, join the voice channel
            # the invoker is in, within that guild
            if (
                interaction.user.voice is not None
                and interaction.user.voice.channel is not None
                and isinstance(interaction.user.voice.channel, discord.VoiceChannel)
            ):
                return interaction.user.voice.channel

        # if invoked from a private message, look at all guilds
        # which have both the bot and the invoking user as a member,
        # find find the first such guild where the user is in a voice
        # channel.
        for guild in interaction.user.mutual_guilds:
            # get member of guild
            member = guild.get_member(interaction.user.id)
            if (
                member is not None
                and member.voice is not None
                and member.voice.channel is not None
                and isinstance(member.voice.channel, discord.VoiceChannel)
            ):
                return member.voice.channel
        return None

    def add_commands(self, tree):
        @discord.app_commands.command(
            name="join_voice",
            description=f"Have {self.persona.ai_name} join the voice "
            + "channel you are in right now.",
        )
        async def join_voice(interaction: discord.Interaction):
            if interaction.user is None:
                await discord_utils.fail_interaction(interaction)

            fancy_logger.get().debug(
                "/join_voice called by user '%s'", interaction.user.name
            )

            voice_channel = self._discover_voice_channel(interaction)
            if voice_channel is None:
                await discord_utils.fail_interaction(
                    interaction, "You must be in a voice channel to use this command"
                )
                return

            # are we already connected to a voice channel?  If so, disconnect
            if self.voice_client is not None:
                fancy_logger.get().debug(
                    "disconnecting from voice channel #%s", self.voice_client.channel
                )
                await self.voice_client.disconnect()
                self.voice_client = None

            await interaction.response.defer(
                ephemeral=True,
                thinking=True,
            )

            try:
                self.voice_client = await voice_channel.connect(
                    cls=voice_client.VoiceClient,
                )
                message = f"Joined voice channel #{voice_channel.name}"
            except discord.DiscordException as err:
                fancy_logger.get().error(
                    "Failed to connect to voice channel #%d: %s", voice_channel.id, err
                )
                message = (
                    f"Failed to connect to voice channel #{voice_channel.name}: {err}"
                )
                return

            await interaction.followup.send(message)

        @discord.app_commands.command(
            name="leave_voice",
            description=f"Have {self.persona.ai_name} leave the "
            + "voice channel it is in.",
        )
        async def leave_voice(interaction: discord.Interaction):
            if interaction.user is None:
                await discord_utils.fail_interaction(interaction)

            fancy_logger.get().debug(
                "/leave_voice called by user: '%s'", interaction.user.name
            )

            # are we already connected to a voice channel?  If so, disconnect
            if self.voice_client is None:
                await discord_utils.fail_interaction(
                    interaction, "Not connected to a voice channel"
                )
                return

            channel = self.voice_client.channel

            fancy_logger.get().debug("leaving voice channel #%s", channel)

            await self.voice_client.disconnect()
            self.voice_client = None

            await interaction.response.send_message(
                "Left voice channel",
                ephemeral=True,
                silent=True,
            )

        fancy_logger.get().debug("Registering audio commands")
        tree.add_command(join_voice)
        tree.add_command(leave_voice)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""
Bot entrypoint.
"""

import asyncio
import signal
import sys
import threading
import typing

import oobabot
from oobabot import discord_utils
from oobabot import fancy_logger
from oobabot import runtime
from oobabot import settings
from oobabot import types
from oobabot import voice_client


class Oobabot:
    """
    Main bot class.  Load settings, creates helper objects,
    and invokes the bot loop.

    Methods:
        constructor: Loads settings from the command line, environment
            variables, and config file.  This config may be changed
            before calling start().
        start: Start the bot.  Blocks until the bot exits.
        stop: Stop the bot.  Blocks until the bot exits.
        is_voice_enabled: Returns True if the bot is configured to
            participate in voice channels.
        current_voice_transcript (property): Returns the current
            voice transcript, or an empty list if the bot is not
            in a voice channel.
        fancy_author_info: Given a Discord user_id, returns a struct
            describing a user's display name, accent color, and icon.
            Bot must be running for this to work.

    Class Methods:
        test_discord_token: Test a discord token to see if it's valid.
            Requires Internet access.
        generate_invite_url: Generate a URL that can be used to invite
            the bot to a server.
    """

    def __init__(
        self,
        cli_args: typing.List[str],
        running_from_cli: bool = False,
    ):
        """
        Initialize the bot, and load settings from the command line,
        environment variables, and config file.  These will be
        available in self.settings.

        self.settings is an :py:class:`oobabot.settings.Settings` object.
        """

        self.runtime: typing.Optional[runtime.Runtime] = None
        # guards access to runtime from multiple threads
        self.runtime_lock = threading.Lock()
        self.settings = settings.Settings()

        try:
            self.settings.load(cli_args, running_from_cli=running_from_cli)
        except settings.SettingsError as err:
            print("\n".join([str(arg) for arg in list(err.args)]), file=sys.stderr)
            raise

        fancy_logger.init_logging(
            level=self.settings.discord_settings.get_str("log_level"),
            running_from_cli=running_from_cli,
        )

    def start(self):
        """
        Start the bot.  Blocks until the bot exits.

        When running inside another process, the bot will exit
        when another thread calls stop().  Otherwise it will
        exit when the user presses Ctrl-C, or the process receives
        a SIGTERM or SIGINT signal.
        """
        fancy_logger.get().info(
            "Starting oobabot, core version %s", oobabot.__version__
        )

        with self.runtime_lock:
            self.runtime = runtime.Runtime(self.settings)
            test_passed = self.runtime.test_connections()
            if not test_passed:
                # test_connections will have logged the error
                self.runtime = None
                return

        asyncio.run(self.runtime.run())

    # pylint: disable=R1732
    def stop(self) -> bool:
        """
        Stop the bot.  Blocks until the bot exits.
        """
        result = self.runtime_lock.acquire(timeout=5)
        if result:
            try:
                if self.runtime is not None:
                    self.runtime.stop()
            finally:
                self.runtime_lock.release()
        else:
            fancy_logger.get().error(
                "Failed to acquire runtime lock, could not shutdown gracefully"
            )
        self.runtime = None
        return result

    # pylint: enable=R1732

    @classmethod
    def test_discord_token(cls, discord_token: str) -> bool:
        """
        Tests a discord token to see if it's valid.  Can be called from any thread.

        Requires Internet connectivity.

        Returns True if it was able to connect with the token, False if it isn't.
        """
        return asyncio.run(discord_utils.test_discord_token(discord_token))

    @classmethod
    def generate_invite_url(cls, discord_token: str) -> str:
        """
        Generates an invite URL for the bot with the given token.

        Can be called from any thread.  Does not require Internet connectivity.
        """
        ai_user_id = discord_utils.get_user_id_from_token(discord_token)
        return discord_utils.generate_invite_url(ai_user_id)

    def is_voice_enabled(self) -> bool:
        """
        Returns True if the bot is configured to participate in voice channels.

        This checks that both:
         a. discrivener is configured
         b. discrivener is installed
        """
        exe_location, model_location = discord_utils.validate_discrivener_locations(
            self.settings.discord_settings.get_str("discrivener_location"),
            self.settings.discord_settings.get_str("discrivener_model_location"),
        )
        return exe_location is not None and model_location is not None

    @property
    def current_voice_transcript(
        self,
    ) -> typing.List["types.VoiceMessage"]:
        """
        If the bot is currently in a voice channel, returns the transcript
        of what's being said.  Otherwise, returns an empty list.
        """
        client = voice_client.VoiceClient.current_instance
        if client is None:
            return []
        transcript = client.current_transcript()
        if transcript is None:
            return []
        return transcript.message_buffer.get()

    def fancy_author_info(self, user_id: int) -> typing.Optional["types.FancyAuthor"]:
        """
        Returns a FancyAuthor object for the given user_id.

        This will only work if the bot is running and connected
        to a voice channel, and then only for users who are
        in that discord server.

        We do this instead of a more general lookup because
        this is what lets us get more information, including
        the server-specific nickname, color, and icon.
        """
        client = voice_client.VoiceClient.current_instance
        if client is None:
            return None
        return discord_utils.author_from_user_id(user_id, client.guild)

    def log_count(self) -> int:
        """
        Returns the current number of times the log has been
        appended to.
        """
        return fancy_logger.recent_logs.changes

    def logs(self) -> typing.List[str]:
        """
        Returns a list of the most recent log messages.
        """
        return fancy_logger.recent_logs.get_all()


def run_cli():
    """
    Run the bot from the command line.  Blocks until the bot exits.

    This is the main entrypoint for the CLI.

    In addition to running the bot, this function also handles
    other CLI commands, such as:
    - help: Print help and exit
    - generate_config: Print a the config file and exit
    - invite_url: Print a URL that can be used to invite the bot
    """

    # create the object and load our settings
    try:
        oobabot = Oobabot(sys.argv[1:], running_from_cli=True)
    except settings.SettingsError:
        sys.exit(1)

    # if we're running as a worker thread in another process,
    # we can't (and shouldn't) register
    def exit_handler(signum, _frame):
        sig_name = signal.Signals(signum).name
        fancy_logger.get().info("Received signal %s, exiting...", sig_name)
        result = oobabot.stop()
        sys.exit(not result)

    signal.signal(signal.SIGINT, exit_handler)
    signal.signal(signal.SIGTERM, exit_handler)

    if oobabot.settings.general_settings.get("help"):
        oobabot.settings.print_help()
        return

    if oobabot.settings.general_settings.get("generate_config"):
        oobabot.settings.write_to_stream(out_stream=sys.stdout)
        if sys.stdout.isatty():
            print(oobabot.settings.META_INSTRUCTION, file=sys.stderr)
        else:
            print("# oobabot: config.yml output successfully", file=sys.stderr)
        return

    if not oobabot.settings.discord_settings.get("discord_token"):
        msg = (
            f"Please set the '{oobabot.settings.DISCORD_TOKEN_ENV_VAR}' "
            + "environment variable to your bot's discord token."
        )
        print(msg, file=sys.stderr)
        sys.exit(1)

    if oobabot.settings.general_settings.get("invite_url"):
        url = oobabot.generate_invite_url(
            oobabot.settings.discord_settings.get_str("discord_token")
        )
        print(url)
        return

    sys.excepthook = fancy_logger.excepthook

    oobabot.start()


def main():
    run_cli()



# -*- coding: utf-8 -*-
"""
Retrieves persona data from a variety of formats.
"""

import json
import re
import typing

import ruamel.yaml as ryaml

from oobabot import fancy_logger


class Persona:
    """
    Handles retrieving persona data from a variety of formats
    """

    # list of keys that, depending on the json/yaml schema, might
    # contain the AI's name.  Take the first one found, in order.
    NAME_KEYS = ["char_name", "name"]

    # list of keys that, depending on the json/yaml schema, might
    # contain the AI's persona.  Take the first one found, in order.
    PERSONA_KEYS = ["char_persona", "description", "context", "personality"]

    def __init__(self, persona_settings: dict) -> None:
        self.ai_name: str = persona_settings["ai_name"]
        self.persona: str = persona_settings["persona"]
        self.wakewords: typing.List[str] = persona_settings["wakewords"].copy()

        # if a json file is specified, load it and have
        # that overwrite everything else
        if "persona_file" in persona_settings:
            filename = persona_settings["persona_file"]
            try:
                self.load_from_file(filename)
            except FileNotFoundError:
                fancy_logger.get().warning(
                    "Could not find persona file: %s",
                    filename,
                )
                return

        # match messages that include any `wakeword`, but not as part of
        # another word
        self.wakeword_patterns = [
            re.compile(rf"\b{wakeword}\b", re.IGNORECASE) for wakeword in self.wakewords
        ]

    def contains_wakeword(self, message: str) -> bool:
        for wakeword_pattern in self.wakeword_patterns:
            if wakeword_pattern.search(message):
                return True
        return False

    def substitute(self, text: str) -> str:
        return text.replace("{{char}}", self.ai_name)

    def load_from_file(self, filename: str):
        if not filename:
            return

        if filename.endswith(".json"):
            self.load_from_json_file(filename)
            return

        if filename.endswith(".yaml"):
            self.load_from_yaml_file(filename)
            return

        if filename.endswith(".txt"):
            self.load_from_text_file(filename)
            return

        fancy_logger.get().warning(
            "Unknown persona file extension (expected .json, or .txt): %s",
            filename,
        )

    def load_from_text_file(self, filename: str):
        with open(filename, "r", encoding="utf-8") as file:
            persona = file.read()
        self.persona = persona

    def load_from_json_file(self, filename: str):
        try:
            with open(filename, "r", encoding="utf-8") as file:
                json_data = json.load(file)

        except json.JSONDecodeError as err:
            fancy_logger.get().warning(
                "Could not parse persona file: %s.  Cause: %s",
                filename,
                err,
            )
            return
        self.load_from_dict(json_data)

    def load_from_yaml_file(self, filename):
        with open(filename, "r", encoding="utf-8") as file:
            yaml = ryaml.YAML(typ="safe")
            try:
                yaml_settings = yaml.load(file)
            except ryaml.YAMLError as err:
                fancy_logger.get().warning(
                    "Could not parse persona file: %s.  Cause: %s",
                    filename,
                    err,
                )
                return
        self.load_from_dict(yaml_settings)

    def load_from_dict(self, json_data: dict):
        for name_key in Persona.NAME_KEYS:
            if name_key in json_data and json_data[name_key]:
                self.ai_name = json_data[name_key]
                break
        for persona_key in Persona.PERSONA_KEYS:
            if persona_key in json_data and json_data[persona_key]:
                self.persona = self.substitute(json_data[persona_key])
                break
        if self.ai_name not in self.wakewords and self.ai_name:
            self.wakewords.append(self.ai_name)



# -*- coding: utf-8 -*-
"""
Contains all runtime state of the bot.
"""

import asyncio
from concurrent import futures
import contextlib
import threading
import typing

import discord

from oobabot import bot_commands
from oobabot import decide_to_respond
from oobabot import discord_bot
from oobabot import fancy_logger
from oobabot import http_client
from oobabot import image_generator
from oobabot import ooba_client
from oobabot import persona
from oobabot import prompt_generator
from oobabot import repetition_tracker
from oobabot import response_stats
from oobabot import sd_client
from oobabot import settings
from oobabot import templates


class OobabotRuntimeError(Exception):
    """
    Raised when the bot encounters an error that it can't recover from.
    """


class Runtime:
    """
    Contains all the runtime state of the bot.  It should be
    created once the configuration is known.
    """

    def __init__(self, settings: settings.Settings):
        # templates used to generate prompts to send to the AI
        # as well as for some UI elements
        self.template_store = templates.TemplateStore(
            settings=settings.template_settings.get_all()
        )

        ########################################################
        # Connect to Oobabooga

        self.ooba_client = ooba_client.OobaClient(
            settings=settings.oobabooga_settings.get_all(),
        )

        ########################################################
        # Connect to Stable Diffusion, if configured

        self.stable_diffusion_client: typing.Optional[
            sd_client.StableDiffusionClient
        ] = None
        sd_settings = settings.stable_diffusion_settings.get_all()
        if sd_settings["stable_diffusion_url"]:
            self.stable_diffusion_client = sd_client.StableDiffusionClient(
                settings=sd_settings,
                magic_model_key=settings.SD_CLIENT_MAGIC_MODEL_KEY,
            )

        ########################################################
        # Bot logic

        self.persona = persona.Persona(
            persona_settings=settings.persona_settings.get_all()
        )

        # decides which messages the bot will respond to
        self.decide_to_respond = decide_to_respond.DecideToRespond(
            discord_settings=settings.discord_settings.get_all(),
            persona=self.persona,
            interrobang_bonus=settings.DECIDE_TO_RESPOND_INTERROBANG_BONUS,
            time_vs_response_chance=settings.TIME_VS_RESPONSE_CHANCE,
        )

        # once we decide to respond, this generates a prompt
        # to send to the AI, given a message history
        self.prompt_generator = prompt_generator.PromptGenerator(
            discord_settings=settings.discord_settings.get_all(),
            oobabooga_settings=settings.oobabooga_settings.get_all(),
            persona=self.persona,
            template_store=self.template_store,
        )

        # tracks of the time spent on responding, success rate, etc.
        self.response_stats = response_stats.AggregateResponseStats(
            fn_get_total_tokens=lambda: self.ooba_client.total_response_tokens
        )

        # generates images, if stable diffusion is configured
        # also includes a UI to regenerate images on demand
        self.image_generator = None
        if self.stable_diffusion_client is not None:
            self.image_generator = image_generator.ImageGenerator(
                ooba_client=self.ooba_client,
                persona_settings=settings.persona_settings.get_all(),
                prompt_generator=self.prompt_generator,
                sd_settings=settings.stable_diffusion_settings.get_all(),
                stable_diffusion_client=self.stable_diffusion_client,
                template_store=self.template_store,
            )

        # if a bot sees itself repeating a message over and over,
        # it will keep doing so forever.  This attempts to fix that.
        # by looking for repeated responses, and deciding how far
        # back in history the bot can see.
        self.repetition_tracker = repetition_tracker.RepetitionTracker(
            repetition_threshold=settings.REPETITION_TRACKER_THRESHOLD
        )

        self.bot_commands = bot_commands.BotCommands(
            decide_to_respond=self.decide_to_respond,
            repetition_tracker=self.repetition_tracker,
            persona=self.persona,
            discord_settings=settings.discord_settings.get_all(),
            template_store=self.template_store,
            ooba_client=self.ooba_client,
            prompt_generator=self.prompt_generator,
        )

        ########################################################
        # Connect to Discord

        self.discord_bot = discord_bot.DiscordBot(
            bot_commands=self.bot_commands,
            decide_to_respond=self.decide_to_respond,
            discord_settings=settings.discord_settings.get_all(),
            ooba_client=self.ooba_client,
            image_generator=self.image_generator,
            persona=self.persona,
            prompt_generator=self.prompt_generator,
            repetition_tracker=self.repetition_tracker,
            response_stats=self.response_stats,
        )

        self.discord_token = settings.discord_settings.get_str("discord_token")

    def test_connections(self) -> bool:
        """
        Tests that we can connect to all services we depend on.
        Does not test Discord connectivity.

        Returns True if all configured connections succeeded,
        False otherwise.
        """

        for client in [self.ooba_client, self.stable_diffusion_client]:
            if client is None:
                continue

            fancy_logger.get().info("%s is at %s", client.service_name, client.base_url)
            try:
                client.test_connection()
                fancy_logger.get().info("Connected to %s!", client.service_name)
            except (ValueError, http_client.OobaHttpClientError) as err:
                fancy_logger.get().warning(
                    "Could not connect to %s server: [%s]",
                    client.service_name,
                    client.base_url,
                )
                fancy_logger.get().warning("Please check the URL and try again.")
                if err.__cause__ is not None:
                    fancy_logger.get().error("Reason: %s", err.__cause__)
                return False
        return True

    async def run(self):
        """
        Opens HTTP connections to oobabooga and stable diffusion,
        then connects to Discord.  Blocks until the bot is stopped.

        Raises OobabotRuntimeError if the bot cannot connect to Discord.
        """

        async with contextlib.AsyncExitStack() as stack:
            for context_manager in [
                self.ooba_client,
                self.stable_diffusion_client,
            ]:
                if context_manager is not None:
                    await stack.enter_async_context(context_manager)

            try:
                fancy_logger.get().info("Connecting to Discord... ")
                await self.discord_bot.start(self.discord_token)
                fancy_logger.get().info("Discord bot exited.")

            except discord.errors.PrivilegedIntentsRequired as err:
                fancy_logger.get().warning("Could not log in to Discord: %s", err)
                fancy_logger.get().warning(
                    "The bot token you provided does not have the required "
                    + "gateway intents.  Did you remember to enable both "
                    + "'SERVER MEMBERS INTENT' and 'MESSAGE CONTENT INTENT' "
                    + "in the bot's settings on Discord?"
                )
                raise OobabotRuntimeError(
                    "Could not log in to Discord: intents not set"
                ) from err

            except discord.LoginFailure as err:
                fancy_logger.get().warning("Could not log in to Discord: %s", err)
                fancy_logger.get().warning("Please check the token and try again.")
                raise OobabotRuntimeError(
                    "Could not log in to Discord: invalid token"
                ) from err

            finally:
                await self.discord_bot.close()
        fancy_logger.get().info("Disconnected from Discord.")

    def stop(self, wait_timeout: float = 5.0) -> None:
        """
        Stops the bot, if it's running.  Safe to be called
        from a separate thread from the one that called run().

        Blocks until the bot is gracefully stopped, or until
        wait_timeout seconds have passed, whichever comes first.
        """

        self.response_stats.write_stat_summary_to_log()

        # if we're on our main thread, then we can just terminate
        # the bot loop directly.  But if we're on a separate thread,
        # then we need to call it from the  discord bot's event loop.
        if threading.current_thread() == threading.main_thread():
            self.discord_bot.loop.stop()
            return

        try:
            # if discord is already stopped, then their .loop is set
            # _MissingSentinel.  So instead check if it's still connected.
            if self.discord_bot.is_closed():
                fancy_logger.get().info("Discord bot already stopped.")
                return

            future = asyncio.run_coroutine_threadsafe(
                self.discord_bot.close(),
                self.discord_bot.loop,
            )
            future.result(timeout=wait_timeout)
        except RuntimeError as err:
            # this can happen if the main application is shutting down
            fancy_logger.get().error("Could not stop Discord bot: %s", err)
            raise OobabotRuntimeError("Discord bot is already stopped.") from err
        except futures.TimeoutError:
            fancy_logger.get().warning(
                "Discord bot did not stop in time, it might be busted."
            )



# -*- coding: utf-8 -*-
"""
Templates for text generated by the bot.  There are two types:
 - templates used to generate prompts for the AI
 - templates used to generate UI messages for the user
"""

import enum
import functools
import textwrap
import typing


@functools.total_ordering
class Templates(enum.Enum):
    """
    Enumeration of all different templates.
    """

    COMMAND_LOBOTOMIZE_RESPONSE = "command_lobotomize_response"

    IMAGE_DETACH = "image_detach"
    IMAGE_CONFIRMATION = "image_confirmation"
    IMAGE_GENERATION_ERROR = "image_generation_error"
    IMAGE_UNAUTHORIZED = "image_unauthorized"

    # prompts to the AI to generate text responses
    PROMPT = "prompt"
    PROMPT_HISTORY_LINE = "prompt_history_line"
    PROMPT_IMAGE_COMING = "prompt_image_coming"

    def __str__(self) -> str:
        return self.value

    def __lt__(self, other: "Templates") -> bool:
        return str(self.value) < str(other.value)


class TemplateToken(str, enum.Enum):
    """
    Enumeration of all tokens used in templates.
    Tokens are variable substitutions into the templates.
    """

    AI_NAME = "AI_NAME"
    PERSONA = "PERSONA"
    IMAGE_COMING = "IMAGE_COMING"
    IMAGE_PROMPT = "IMAGE_PROMPT"
    MESSAGE_HISTORY = "MESSAGE_HISTORY"
    USER_MESSAGE = "USER_MESSAGE"
    USER_NAME = "USER_NAME"

    def __str__(self):
        return "{" + self.value + "}"


class TemplateStore:
    """
    Data object storing all template definitions and default values.
    """

    # mapping of template names to tokens allowed in that template
    #  key: template name
    #  value: tuple of (list of tokens, description, is_an_ai_prompt)
    TEMPLATES: typing.Dict[
        Templates, typing.Tuple[typing.List[TemplateToken], str, bool]
    ] = {
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: (
            [
                TemplateToken.AI_NAME,
                TemplateToken.USER_NAME,
            ],
            "Displayed in Discord after a successful /lobotomize command.  "
            + "Both the discord users and the bot AI will see this message.",
            True,
        ),
        Templates.PROMPT: (
            [
                TemplateToken.AI_NAME,
                TemplateToken.IMAGE_COMING,
                TemplateToken.MESSAGE_HISTORY,
                TemplateToken.PERSONA,
            ],
            "The main prompt sent to Oobabooga to generate a response from "
            + "the bot AI.  The AI's reply to this prompt will be sent to "
            + "discord as the bot's response.",
            True,
        ),
        Templates.PROMPT_HISTORY_LINE: (
            [
                TemplateToken.USER_MESSAGE,
                TemplateToken.USER_NAME,
            ],
            "Part of the AI response-generation prompt, this is used to "
            + "render a single line of chat history.  A list of these, "
            + "one for each past chat message, will become {MESSAGE_HISTORY} "
            + "and inserted into the main prompt",
            True,
        ),
        Templates.PROMPT_IMAGE_COMING: (
            [
                TemplateToken.AI_NAME,
            ],
            "Part of the AI response-generation prompt, this is used to "
            + "inform the AI that it is in the process of generating an "
            + "image.",
            True,
        ),
        Templates.IMAGE_DETACH: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when the user selects to discard an image "
            + "that Stable Diffusion had generated.",
            False,
        ),
        Templates.IMAGE_CONFIRMATION: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when an image is first generated from "
            + "Stable Diffusion.  This should prompt the user to either "
            + "save or discard the image.",
            False,
        ),
        Templates.IMAGE_GENERATION_ERROR: (
            [
                TemplateToken.IMAGE_PROMPT,
                TemplateToken.USER_NAME,
            ],
            "Shown in Discord when the we could not contact Stable Diffusion "
            + "to generate an image.",
            False,
        ),
        Templates.IMAGE_UNAUTHORIZED: (
            [TemplateToken.USER_NAME],
            "Shown in Discord privately to a user if they try to regenerate "
            "an image that was requested by someone else.",
            False,
        ),
    }

    DEFAULT_TEMPLATES: typing.Dict[Templates, str] = {
        Templates.PROMPT: textwrap.dedent(
            """
            You are in a chat room with multiple participants.
            Below is a transcript of recent messages in the conversation.
            Write the next one to three messages that you would send in this
            conversation, from the point of view of the participant named
            {AI_NAME}.

            {PERSONA}

            All responses you write must be from the point of view of
            {AI_NAME}.
            ### Transcript:
            {MESSAGE_HISTORY}
            {IMAGE_COMING}
            """
        ),
        Templates.PROMPT_HISTORY_LINE: textwrap.dedent(
            """
            {USER_NAME} says:
            {USER_MESSAGE}

            """
        ),
        Templates.PROMPT_IMAGE_COMING: textwrap.dedent(
            """
            {AI_NAME}: is currently generating an image, as requested.
            """
        ),
        Templates.IMAGE_DETACH: textwrap.dedent(
            """
            {USER_NAME} asked for an image with the prompt:
                '{IMAGE_PROMPT}'
            ...but couldn't find a suitable one.
            """
        ),
        Templates.IMAGE_CONFIRMATION: textwrap.dedent(
            """
            {USER_NAME}, is this what you wanted?
            If no choice is made, this message will 💣 self-destruct 💣 in 3 minutes.
            """
        ),
        Templates.IMAGE_GENERATION_ERROR: textwrap.dedent(
            """
            Something went wrong generating your image.  Sorry about that!
            """
        ),
        Templates.IMAGE_UNAUTHORIZED: textwrap.dedent(
            """
            Sorry, only {USER_NAME} can press the buttons.
            """
        ),
        Templates.COMMAND_LOBOTOMIZE_RESPONSE: textwrap.dedent(
            """
            Ummmm... what were we talking about?
            """
        ),
    }

    def __init__(self, settings: dict):
        self.templates: typing.Dict[Templates, TemplateMessageFormatter] = {}
        for template, (tokens, purpose, is_ai_prompt) in self.TEMPLATES.items():
            template_name = str(template)
            template_fmt = settings[template_name]
            if template_fmt is None:
                raise ValueError(f"Template {template_name} has no default format")
            self.add_template(template, template_fmt, tokens, purpose, is_ai_prompt)

    def add_template(
        self,
        template_name: Templates,
        format_str: str,
        allowed_tokens: typing.List[TemplateToken],
        purpose: str,
        is_ai_prompt: bool,
    ):
        self.templates[template_name] = TemplateMessageFormatter(
            template_name,
            format_str,
            allowed_tokens,
            purpose,
            is_ai_prompt,
        )

    def format(
        self, template_name: Templates, format_args: typing.Dict[TemplateToken, str]
    ) -> str:
        return self.templates[template_name].format(format_args)


class TemplateMessageFormatter:
    """
    Validates that templates are safe to run string.format() on, and
    runs string.format()
    """

    def __init__(
        self,
        template_name: Templates,
        template: str,
        allowed_tokens: typing.List[TemplateToken],
        purpose: str,
        is_ai_prompt: bool,
    ):
        self._validate_format_string(template_name, template, allowed_tokens)
        self.template_name = template_name
        self.template = template
        self.allowed_tokens = allowed_tokens
        self.purpose = purpose
        self.is_ai_prompt = is_ai_prompt

    def __str__(self):
        return self.template

    def format(self, format_args: typing.Dict[TemplateToken, str]) -> str:
        return self.template.format(**format_args)

    @staticmethod
    def _validate_format_string(
        template_name: Templates,
        format_str: str,
        allowed_args: typing.List[TemplateToken],
    ):
        def find_all_ch(string: str, char: str) -> typing.Generator[int, None, None]:
            # find all indices of ch in s
            for i, letter in enumerate(string):
                if letter == char:
                    yield i

        # raises if fmt_string contains any args not in allowed_args
        allowed_close_brace_indices: typing.Set[int] = set()

        for open_brace_idx in find_all_ch(format_str, "{"):
            for allowed_arg in allowed_args:
                idx_end = open_brace_idx + len(allowed_arg) + 1
                next_substr = format_str[open_brace_idx : idx_end + 1]
                if next_substr == "{" + allowed_arg + "}":
                    allowed_close_brace_indices.add(idx_end)
                    break
            else:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )
        for close_brace_idx in find_all_ch(format_str, "}"):
            if close_brace_idx not in allowed_close_brace_indices:
                raise ValueError(
                    f"invalid template: {template_name} contains "
                    + f"an argument not in {allowed_args}"
                )



# -*- coding: utf-8 -*-
"""
Client for generating images using the AUTOMATIC1111
API.  Takes a prompt and returns image binary data in PNG format.
"""

import asyncio
import base64
import re
import time
import typing

import aiohttp

from oobabot import fancy_logger
from oobabot import http_client

# todo: response stats for SD client


def _find_substring_in_dict(
    desired_val: str, search_list: typing.List
) -> typing.Optional[str]:
    desired_val = desired_val.lower()
    for value in search_list:
        if value.lower() in desired_val:
            return value
    return None


class StableDiffusionClient(http_client.SerializedHttpClient):
    """
    Purpose: Client for generating images using the AUTOMATIC1111
    API.  Takes a prompt and returns image binary data in PNG format.
    """

    SERVICE_NAME = "Stable Diffusion"
    STABLE_DIFFUSION_API_URI_PATH: str = "/sdapi/v1/"

    SAMPLER_KEY = "sampler_name"
    SAMPLER_KEY_ALIAS = "sampler"

    API_COMMAND_URLS = {
        "get_samplers": STABLE_DIFFUSION_API_URI_PATH + "samplers",
        # get_samplers: GET only
        #   returns a list of samplers which we can use
        #   in text2img
        "options": STABLE_DIFFUSION_API_URI_PATH + "options",
        # options: GET and POST
        #   retrieves (GET) and sets (POST) global options
        #   for the server.  This is how we set the checkpoint
        #   and also a few privacy-related fields.
        "progress": STABLE_DIFFUSION_API_URI_PATH + "progress",
        # progress: GET only
        #   returns the progress of the current image generation
        "txt2img": STABLE_DIFFUSION_API_URI_PATH + "txt2img",
        # txt2img: POST only
        #   takes a prompt, generates an image and returns it
        "sd-models": STABLE_DIFFUSION_API_URI_PATH + "sd-models",
        # sd-models: GET only
        # [
        #   {
        #     "title": "Anything-V3.0.ckpt [812cd9f9d9]",
        #     "model_name": "Anything-V3.0",
        #     "hash": "812cd9f9d9",
        #     "sha256": "812cd9f9d9a0cb62aaad605173fd64dea1....",
        #     "filename": "...long..path.../models/Stable-diffusion/Anything-V3.0.ckpt",
        #     "config": null
        #   },
        # ...
        # ]
    }

    def __init__(
        self,
        settings: typing.Dict[str, typing.Any],
        magic_model_key: str,
    ):
        super().__init__(self.SERVICE_NAME, settings["stable_diffusion_url"])

        self.extra_prompt_text = settings["extra_prompt_text"]
        self.request_params = settings["request_params"]

        # lower-case all keys in request_params
        self.request_params = {k.lower(): v for k, v in self.request_params.items()}
        self.sd_models = []
        self.sd_samplers = []

        self.user_override_params = {}
        # ensure that each customizable param is in the request_params
        for param in settings["user_override_params"]:
            param = param.lower()
            if param not in self.request_params:
                fancy_logger.get().warning(
                    "Stable Diffusion:  customizable param '%s' not in request_params."
                    + "  Ignoring setting.",
                    param,
                )
                continue
            # store the type of the param, so we can validate user input later
            self.user_override_params[param] = type(self.request_params[param])

        self.magic_model_key = magic_model_key.lower()

        # when we're in a "age restricted" channel, we'll swap
        # the "negative_prompt" in the request_params with this
        # value.  Otherwise we'll use the one already in
        # request_params["negative_prompt"]
        self.negative_prompt_nsfw = self.request_params.pop("negative_prompt_nsfw", "")

    DEFAULT_OPTIONS = {
        #
        # - "enable_pnginfo"
        #   by default, this will store the parameters used to generate the image.
        #   For instance:
        #         Parameters: zombie taylor swift pinup movie posterSteps: 30,
        #           Sampler: DPM++ 2M Karras, CFG scale: 7, Seed: 1317257172,
        #           Size: 768x768, Model hash: 9aba26abdf, Model: Deliberate-2.0-15236
        #   We want to disable this, for a few reasons:
        #   - it's a privacy concern, since it's not clear to the user what
        #     is happening
        #   - it will include our negative prompts, which are actually bad
        #     phrases that represent content we DON'T want to generate.  If
        #     we upload these same strings to Discord, they might mistakenly
        #     flag the content as violating their TOS.  Even though the only
        #     reason those strings are used is to prevent that content.
        "enable_pnginfo": False,
        #
        # - "do_not_add_watermark"
        #    Similar to the above, we don't want to add a watermark to the
        #    generated image, just since it looks bad.  Also, it can leak
        #    the version of the SD software we are using, which isn't a good
        #    idea for security reasons.
        "do_not_add_watermark": True,
        #
        # - "samples_format"
        #   this is the format used for all image generation.
        "samples_format": "png",
    }

    async def set_options(self):
        url = self.API_COMMAND_URLS["options"]

        # first, get the current options.  We do this for two
        # reasons;
        #  - show the user exactly what we changed, if anything
        #  - some versions of the API don't support the
        #    "do_not_add_watermark" option, so we need to
        #    check if it's there before we try to set it
        #
        current_options = None
        async with self._get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            current_options = await response.json()

        options_to_set = {}
        for option_name, option_value in self.DEFAULT_OPTIONS.items():
            if option_name not in current_options:
                continue
            if option_value == current_options[option_name]:
                continue
            options_to_set[option_name] = option_value
            fancy_logger.get().info(
                "Stable Diffusion:  changing option '%s' from to '%s' to '%s'",
                option_name,
                current_options[option_name],
                option_value,
            )

        if not options_to_set:
            fancy_logger.get().debug(
                "Stable Diffusion: Options are already set correctly, no changes made."
            )
            return

        async with self._get_session().post(url, json=options_to_set) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            await response.json()

    async def _call_and_extract_field(
        self, command: str, field: str
    ) -> typing.List[str]:
        url = self.API_COMMAND_URLS[command]
        async with self._get_session().get(url) as response:
            if response.status != 200:
                raise http_client.OobaHttpClientError(response)
            response = await response.json()
            values = [str(value[field]) for value in response]
            return values

    async def get_samplers(self) -> typing.List[str]:
        return await self._call_and_extract_field("get_samplers", "name")

    async def get_models(self) -> typing.List[str]:
        return await self._call_and_extract_field("sd-models", "model_name")

    SD_DELIMITER = "="

    def _find_model(self, desired_model: str) -> typing.Optional[str]:
        return _find_substring_in_dict(desired_model, self.sd_models)

    def _find_sampler(self, desired_sampler: str) -> typing.Optional[str]:
        return _find_substring_in_dict(desired_sampler, self.sd_samplers)

    def _to_key_value_pair(
        self, word: str
    ) -> typing.Optional[typing.Tuple[str, typing.Any]]:
        if self.SD_DELIMITER not in word:
            return None

        keyword_pair = word.split(self.SD_DELIMITER, 1)
        if len(keyword_pair) < 2:
            return None
        key, val = keyword_pair

        # lowercase the key, so we can match it against the
        # user_override_params dict
        key = key.lower()

        # magical alias for "negative_prompt"
        if key == "np":
            key = "negative_prompt"

        if key == self.SAMPLER_KEY_ALIAS:
            key = self.SAMPLER_KEY

        if key not in self.user_override_params:
            return None

        # try to remove any quotes around the value
        if val.startswith('"') and val.endswith('"'):
            val = val[1:-1]

        try:
            val_type = self.user_override_params[key]
            if val_type == bool:
                if val.lower in ("true", "yes", "y"):
                    val = True
                elif val.lower in ("false", "no", "n"):
                    val = False
            elif val_type == int:
                val = int(val)
            elif val_type == float:
                val = float(val)
        except ValueError:
            return None

        return (key, val)

    def update_model_and_sampler(self, params: typing.Dict[str, typing.Any]):
        """
        Model and Sampler are special since we have a list of known acceptable
        values.  If the user specifies a value that is not in the list, we
        ignore it.  If the user specifies a value that is in the list, we
        override the default.

        Also, "sampler" is an alias for "sampler_name", and "model" is an
        alias for what should appear in "override_settings.sd_model_checkpoint".

        Finally, we allow the user to specify a substring of the model or
        sampler name, and we will match the first one we find.  This is
        useful for when the user doesn't know the exact name of the model
        or sampler, but knows a substring of it.
        """
        desired_model = params.pop(self.magic_model_key, None)
        if desired_model is not None:
            model = self._find_model(desired_model)
            if model is not None:
                if "override_settings" not in params:
                    params["override_settings"] = {}
                params["override_settings"]["sd_model_checkpoint"] = model
                fancy_logger.get().debug(
                    "Stable Diffusion: per user request, setting model to '%s'", model
                )
            else:
                fancy_logger.get().debug(
                    "Stable Diffusion: unavailable model requested.  "
                    + "If newly added, restart oobabot: '%s'",
                    desired_model,
                )

        # the proper key is "sampler_name", but we also allow
        # "sampler" for convenience.  Move the value, if any, over now.
        desired_sampler = params.pop(self.SAMPLER_KEY_ALIAS, None)
        if desired_sampler is not None:
            params[self.SAMPLER_KEY] = desired_sampler

        desired_sampler = params.pop(self.SAMPLER_KEY, None)
        if desired_sampler is not None:
            sampler = self._find_sampler(desired_sampler)
            if sampler is not None:
                params[self.SAMPLER_KEY] = sampler
            else:
                fancy_logger.get().debug(
                    "Stable Diffusion: unavailable sampler requested.  "
                    + "If newly added, restart oobabot: '%s'",
                    desired_sampler,
                )

    # this is intended to split the prompt into words, but
    # also to handle quoted strings.  For instance, if the
    # prompt is:
    #   "zombie taylor swift" pinup movie poster
    # then we want to split it into:
    #   ["zombie taylor swift", "pinup", "movie", "poster"]
    #
    SD_PARAM_SPLIT_REGEX = re.compile(r'(?:".*?"|\S)+')

    def update_params(self, prompt: str, params: typing.Dict[str, typing.Any]) -> str:
        """
        Updates the request parameters included in the given prompt,
        and then returns the remaining prompt.
        """
        remaining_prompt = ""
        for word in self.SD_PARAM_SPLIT_REGEX.findall(prompt):
            key_val_pair = self._to_key_value_pair(word)
            if key_val_pair is None:
                remaining_prompt += word + " "
                continue
            key, val = key_val_pair

            fancy_logger.get().debug(
                "Stable Diffusion: per user request, setting '%s' to '%s'", key, val
            )
            params[key] = val

        self.update_model_and_sampler(params)

        return remaining_prompt.strip()

    def generate_image(
        self,
        prompt: str,
        is_channel_nsfw: bool = False,
    ) -> "asyncio.Task[bytes]":
        """
        Generate an image from a prompt.
        Args:
            prompt: The prompt to generate an image from.
            is_channel_nsfw: Whether the channel is NSFW.
            this will change the negative prompt.
        Returns:
            The image as bytes.
        Raises:
            OobaHttpClientError, if the request fails.
        """
        request = self.request_params.copy()

        if is_channel_nsfw:
            request["negative_prompt"] = self.negative_prompt_nsfw

        # extract any allowed user overrides from the prompt
        # and add them to the request dict.  The remaining
        # prompt is returned.
        request["prompt"] = self.update_params(prompt, request)

        if self.extra_prompt_text:
            request["prompt"] += ", " + self.extra_prompt_text

        async def do_post() -> bytes:
            fancy_logger.get().debug(
                "Stable Diffusion: Image request (nsfw: %r): %s",
                is_channel_nsfw,
                request,
            )
            start_time = time.time()

            async with self._get_session().post(
                self.API_COMMAND_URLS["txt2img"],
                json=request,
            ) as response:
                if response.status != 200:
                    raise http_client.OobaHttpClientError(response)
                duration = time.time() - start_time
                json_body = await response.json()
                image_bytes = base64.b64decode(json_body["images"][0])
                fancy_logger.get().debug(
                    "Stable Diffusion: Image generated, %d bytes in %.2f seconds",
                    len(image_bytes),
                    duration,
                )
                return image_bytes

        # works around aiohttp being bad
        async def do_post_with_retry() -> bytes:
            tries = 0
            while True:
                try:
                    return await do_post()
                except (aiohttp.ClientError, aiohttp.ClientOSError) as err:
                    retry = False
                    if err.__cause__ is ConnectionResetError:
                        retry = True
                    if isinstance(err, aiohttp.ClientOSError) and 104 == err.errno:
                        retry = True
                    if tries > 2:
                        retry = False
                    if not retry:
                        raise

                    fancy_logger.get().warning(
                        "Stable Diffusion: Connection reset error: %s, "
                        + "retrying in 1 second",
                        err,
                    )
                    await asyncio.sleep(1)
                    tries += 1

        return asyncio.create_task(do_post_with_retry())

    async def _setup(self):
        await self.set_options()
        self.sd_models = await self.get_models()
        self.sd_samplers = await self.get_samplers()
        fancy_logger.get().info(
            "Stable Diffusion: Available models: %s", ", ".join(self.sd_models)
        )
        # log a warning if the default model is not available
        desired_model = self.request_params.get(self.magic_model_key, None)
        if desired_model:
            model = self._find_model(desired_model)
            if model is None:
                fancy_logger.get().warning(
                    "Stable Diffusion: Desired default model '%s' not available.",
                    desired_model,
                )
            else:
                fancy_logger.get().debug(
                    "Stable Diffusion: Using default model '%s'", desired_model
                )

        fancy_logger.get().info(
            "Stable Diffusion: Available samplers: %s", ", ".join(self.sd_samplers)
        )
        # log a warning if the default sampler is not available
        desired_sampler = self.request_params.get(
            self.SAMPLER_KEY_ALIAS, self.request_params.get(self.SAMPLER_KEY, None)
        )
        if desired_sampler:
            sampler = self._find_sampler(desired_sampler)
            if sampler is None:
                fancy_logger.get().warning(
                    "Stable Diffusion: Desired default sampler '%s' not available.",
                    desired_sampler,
                )
            else:
                fancy_logger.get().debug(
                    "Stable Diffusion: Using default sampler '%s'", desired_sampler
                )

        fancy_logger.get().debug(
            "Stable Diffusion: Using negative prompt: %s...",
            str(self.request_params.get("negative_prompt", ""))[:20],
        )
        if self.extra_prompt_text:
            fancy_logger.get().debug(
                "Stable Diffusion: Bot will append to every prompt: '%s'",
                self.extra_prompt_text,
            )
        if 0 == len(self.user_override_params):
            fancy_logger.get().debug(
                "Stable Diffusion: Users cannot override any SD params"
            )
        else:
            fancy_logger.get().debug(
                "Stable Diffusion: Users may override: %s",
                ", ".join(self.user_override_params.keys()),
            )
